@@ -141,6 +141,12 @@ ulong_to_pointer(unsigned long value)
 	return (void *) value;
 }
 
+static inline G_GNUC_CONST WARN_UNUSED_RESULT ALWAYS_INLINE void *
+long_to_pointer(long value)
+{
+	return (void *) value;
+}
+
 static inline G_GNUC_CONST WARN_UNUSED_RESULT ALWAYS_INLINE void *
 size_to_pointer(size_t value)
 {
@@ -153,6 +159,12 @@ pointer_to_ulong(const void *p)
 	return (unsigned long) p;
 }
 
+static inline G_GNUC_CONST WARN_UNUSED_RESULT ALWAYS_INLINE long
+pointer_to_long(const void *p)
+{
+	return (long) p;
+}
+
 static inline G_GNUC_CONST WARN_UNUSED_RESULT ALWAYS_INLINE size_t
 pointer_to_size(const void *p)
 {
@@ -183,6 +195,18 @@ pointer_to_int(const void *p)
 	return pointer_to_uint(p);
 }
 
+static inline G_GNUC_CONST WARN_UNUSED_RESULT ALWAYS_INLINE void *
+bool_to_pointer(bool value)
+{
+	return ulong_to_pointer(value);
+}
+
+static inline G_GNUC_CONST WARN_UNUSED_RESULT ALWAYS_INLINE bool
+pointer_to_bool(const void *p)
+{
+	return p != NULL;
+}
+
 typedef void (*func_ptr_t)();
 
 static inline G_GNUC_CONST WARN_UNUSED_RESULT ALWAYS_INLINE func_ptr_t
