@@ -36,6 +36,19 @@
 
 #include "config.h"
 
+/**
+ * Undefining this symbol will cause gkt-gnutella to avoid using uninitialized
+ * values on the stack to collect initial entropy for random number seeds,
+ * and will prevent errors about "uninitialized value usage" when running under
+ * valgrind.
+ *
+ * By default, this symbol is defined because reading the uninitialized bytes
+ * results in undefined behaviour that precisely adds uncertainety, aka entropy.
+ */
+#if 1
+#define ALLOW_UNINIT_VALUES
+#endif
+
 #ifdef MINGW32
 /*
  * Must set WINVER before including ANY include on recent MinGW installations
@@ -340,10 +353,6 @@ typedef int socket_fd_t;
 #include "lib/regex.h"
 #endif	/* HAS_REGCOMP */
 
-#ifdef USE_GLIB1
-typedef void (*GCallback) (void);
-#endif	/* USE_GLIB1 */
-
 #ifdef USE_GLIB2
 #undef G_STRLOC			/* Want our version */
 #undef G_STRFUNC		/* Version from glib uses __PRETTY_FUNCTION__ */
@@ -569,9 +578,19 @@ typedef void (*GCallback) (void);
  * or seems to be unused.
  */
 #if defined(HASATTRIBUTE) && HAS_GCC(3, 1)
-#define KEEP_FUNCTION __attribute__((__used__))
+#define G_GNUC_USED __attribute__((__used__))
 #else /* GCC < 3.1 || !GCC */
-#define KEEP_FUNCTION
+#define G_GNUC_USED
+#endif
+
+/*
+ * Let the compiler know that the function may be unused, hence it should
+ * not emit any warning about it.
+ */
+#if defined(HASATTRIBUTE) && HAS_GCC(3, 1)
+#define G_GNUC_UNUSED __attribute__((__unused__))
+#else	/* GCC < 3.1 */
+#define G_GNUC_UNUSED
 #endif
 
 /*
@@ -581,7 +600,7 @@ typedef void (*GCallback) (void);
  * does not suppress this warning.
  */
 #define IGNORE_RESULT(x) \
-	G_STMT_START { switch (0 != (x)) { default: ; } }  G_STMT_END
+	G_STMT_START { if (0 != (x)) {} }  G_STMT_END
 
 /*
  * Functions using this attribute cause a warning if the variable
@@ -761,30 +780,6 @@ typedef void (*GCallback) (void);
  */
 #define SIGN(x) (G_UNLIKELY((x) == 0) ? 0 : (x) > 0 ? 1 : (-1))
 
-/**
- * UINT32_SWAP_CONSTANT() byte-swaps a 32-bit word, preferrably a constant.
- * If the value is a variable, use UINT32_SWAP().
- */
-#define UINT32_SWAP_CONSTANT(x_) ((uint32) (		 \
-    (((uint32) (x_) & (uint32) 0x000000ffU) << 24) | \
-    (((uint32) (x_) & (uint32) 0x0000ff00U) <<  8) | \
-    (((uint32) (x_) & (uint32) 0x00ff0000U) >>  8) | \
-    (((uint32) (x_) & (uint32) 0xff000000U) >> 24)))
-
-/**
- * UINT32_SWAP() byte-swaps a 32-bit word.
- *
- * Avoid using glib's GUINT32_SWAP_LE_BE(): it triggers compile-time
- * warnings on a wrong __asm__ statement with glib 1.2.  This version
- * should be as efficient as the one defined by glib.
- */
-#ifdef HAS_BUILTIN_BSWAP32
-#define UINT32_SWAP(x_) \
-	(IS_CONSTANT(x_) ? UINT32_SWAP_CONSTANT(x_) : __builtin_bswap32(x_))
-#else
-#define UINT32_SWAP(x_) UINT32_SWAP_CONSTANT(x_)
-#endif
-
 /**
  * Byte-wise swap two items of specified size.
  */
@@ -914,6 +909,22 @@ ngettext_(const gchar *msg1, const gchar *msg2, ulong n)
  */
 #define ZERO(x)		memset((x), 0, sizeof *(x))
 
+/**
+ * Generate argument list for the address of `x' and its size, so that we can
+ * process the content of that variable.
+ *
+ * The aim is to prevent typos and make sure the two arguments are in sync.
+ */
+#define VARLEN(x)		&(x), sizeof(x)
+
+/**
+ * Generate argument list for the address pointed at by `x' and its size, so
+ * that we can process the structure content, pointed at by that variable.
+ *
+ * The aim is to prevent typos and make sure the two arguments are in sync.
+ */
+#define PTRLEN(x)		(x), sizeof *(x)
+
 /*
  * Support for alloca().
  */
@@ -932,6 +943,17 @@ ngettext_(const gchar *msg1, const gchar *msg2, ulong n)
 #endif	/* alloca */
 #endif	/* HAS_GCC(3, 0) */
 
+/**
+ * Let the program use sigprocmask() but remap it to pthread_sigmask()
+ * to ensure that we manipulate the thread's signal mask only.
+ *
+ * On linux, sigprocmask() always manipulates the thread's signal mask, but
+ * this is not guaranteed by POSIX.
+ */
+#if defined(HAS_SIGPROCMASK) && defined(I_PTHREAD)
+#define sigprocmask(h,s,o)	pthread_sigmask((h), (s), (o))
+#endif
+
 /*
  * Common inclusions, likely to be needed by most files.
  */
