@@ -41,10 +41,12 @@
 #include "if/core/wrap.h"		/* For wrapped_io_t */
 #include "if/gnet_property_priv.h"
 
-#include "lib/glib-missing.h"
+#include "lib/entropy.h"
 #include "lib/halloc.h"
 #include "lib/inputevt.h"
 #include "lib/parse.h"
+#include "lib/plist.h"
+#include "lib/pslist.h"
 #include "lib/stringify.h"
 #include "lib/vmm.h"
 #include "lib/walloc.h"
@@ -114,8 +116,8 @@ enum bsched_magic {
 struct bsched {
 	enum bsched_magic magic;
 	tm_t last_period;			/**< Last time we ran our period */
-	GList *sources;				/**< List of bio_source_t */
-	GSList *stealers;			/**< List of bsched_t stealing bw */
+	plist_t *sources;			/**< List of bio_source_t */
+	pslist_t *stealers;			/**< List of bsched_t stealing bw */
 	char *name;					/**< Name, for tracing purposes */
 	int count;					/**< Amount of sources */
 	int type;					/**< Scheduling type */
@@ -149,9 +151,9 @@ struct bsched {
 
 static bsched_t *bws_set[NUM_BSCHED_BWS];
 
-static GSList *bws_list = NULL;
-static GSList *bws_out_list = NULL;
-static GSList *bws_in_list = NULL;
+static pslist_t *bws_list = NULL;
+static pslist_t *bws_out_list = NULL;
+static pslist_t *bws_in_list = NULL;
 static int bws_out_ema = 0;
 static int bws_in_ema = 0;
 
@@ -242,11 +244,11 @@ bsched_get(bsched_bws_t bws)
 static void
 bsched_free(bsched_t *bs)
 {
-	GList *iter;
+	plist_t *iter;
 
 	bsched_check(bs);
 	
-	for (iter = bs->sources; iter; iter = g_list_next(iter)) {
+	PLIST_FOREACH(bs->sources, iter) {
 		bio_source_t *bio = iter->data;
 
 		bio_check(bio);
@@ -254,8 +256,8 @@ bsched_free(bsched_t *bs)
 		bio->bws = BSCHED_BWS_INVALID;	/* Mark orphan source */
 	}
 
-	gm_list_free_null(&bs->sources);
-	gm_slist_free_null(&bs->stealers);
+	plist_free_null(&bs->sources);
+	pslist_free_null(&bs->stealers);
 	HFREE_NULL(bs->name);
 	bs->magic = 0;
 	WFREE(bs);
@@ -358,7 +360,7 @@ bsched_add_stealer(bsched_bws_t bws, bsched_bws_t bws_stealer)
 	stealer = bsched_get(bws_stealer);
 	g_assert((bs->flags & BS_F_RW) == (stealer->flags & BS_F_RW));
 
-	bs->stealers = g_slist_prepend(bs->stealers, stealer);
+	bs->stealers = pslist_prepend(bs->stealers, stealer);
 }
 
 /**
@@ -369,7 +371,7 @@ bsched_reset_stealers(bsched_t *bs)
 {
 	bsched_check(bs);
 
-	gm_slist_free_null(&bs->stealers);
+	pslist_free_null(&bs->stealers);
 }
 
 /**
@@ -388,10 +390,10 @@ bsched_dht_cross_stealing(void)
 G_GNUC_COLD void
 bsched_config_steal_http_gnet(void)
 {
-	GSList *iter;
+	pslist_t *iter;
 
-	for (iter = bws_list; iter; iter = g_slist_next(iter)) {
-		bsched_bws_t bws = GPOINTER_TO_UINT(iter->data);
+	PSLIST_FOREACH(bws_list, iter) {
+		bsched_bws_t bws = pointer_to_uint(iter->data);
 		bsched_reset_stealers(bsched_get(bws));
 	}
 
@@ -438,9 +440,9 @@ bsched_config_steal_http_gnet(void)
 G_GNUC_COLD void
 bsched_config_steal_gnet(void)
 {
-	GSList *iter;
+	pslist_t *iter;
 
-	for (iter = bws_list; iter; iter = g_slist_next(iter)) {
+	PSLIST_FOREACH(bws_list, iter) {
 		bsched_bws_t bws = GPOINTER_TO_UINT(iter->data);
 		bsched_reset_stealers(bsched_get(bws));
 	}
@@ -501,64 +503,64 @@ bsched_early_init(void)
 	bws_set[BSCHED_BWS_DHT_IN] = bsched_make("DHT in",
 		BS_T_STREAM, BS_F_READ, 0, 1000);
 
-	bws_list = g_slist_prepend(bws_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_LOOPBACK_IN));
-	bws_list = g_slist_prepend(bws_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_PRIVATE_IN));
-	bws_list = g_slist_prepend(bws_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_GLIN));
-	bws_list = g_slist_prepend(bws_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_GIN));
-	bws_list = g_slist_prepend(bws_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_GIN_UDP));
-	bws_list = g_slist_prepend(bws_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_IN));
-	bws_list = g_slist_prepend(bws_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_DHT_IN));
-	bws_list = g_slist_prepend(bws_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_LOOPBACK_OUT));
-	bws_list = g_slist_prepend(bws_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_PRIVATE_OUT));
-	bws_list = g_slist_prepend(bws_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_GLOUT));
-	bws_list = g_slist_prepend(bws_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_GOUT));
-	bws_list = g_slist_prepend(bws_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_GOUT_UDP));
-	bws_list = g_slist_prepend(bws_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_OUT));
-	bws_list = g_slist_prepend(bws_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_DHT_OUT));
+	bws_list = pslist_prepend(bws_list,
+						uint_to_pointer(BSCHED_BWS_LOOPBACK_IN));
+	bws_list = pslist_prepend(bws_list,
+						uint_to_pointer(BSCHED_BWS_PRIVATE_IN));
+	bws_list = pslist_prepend(bws_list,
+						uint_to_pointer(BSCHED_BWS_GLIN));
+	bws_list = pslist_prepend(bws_list,
+						uint_to_pointer(BSCHED_BWS_GIN));
+	bws_list = pslist_prepend(bws_list,
+						uint_to_pointer(BSCHED_BWS_GIN_UDP));
+	bws_list = pslist_prepend(bws_list,
+						uint_to_pointer(BSCHED_BWS_IN));
+	bws_list = pslist_prepend(bws_list,
+						uint_to_pointer(BSCHED_BWS_DHT_IN));
+	bws_list = pslist_prepend(bws_list,
+						uint_to_pointer(BSCHED_BWS_LOOPBACK_OUT));
+	bws_list = pslist_prepend(bws_list,
+						uint_to_pointer(BSCHED_BWS_PRIVATE_OUT));
+	bws_list = pslist_prepend(bws_list,
+						uint_to_pointer(BSCHED_BWS_GLOUT));
+	bws_list = pslist_prepend(bws_list,
+						uint_to_pointer(BSCHED_BWS_GOUT));
+	bws_list = pslist_prepend(bws_list,
+						uint_to_pointer(BSCHED_BWS_GOUT_UDP));
+	bws_list = pslist_prepend(bws_list,
+						uint_to_pointer(BSCHED_BWS_OUT));
+	bws_list = pslist_prepend(bws_list,
+						uint_to_pointer(BSCHED_BWS_DHT_OUT));
 
-	bws_in_list = g_slist_prepend(bws_in_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_LOOPBACK_IN));
-	bws_in_list = g_slist_prepend(bws_in_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_PRIVATE_IN));
-	bws_in_list = g_slist_prepend(bws_in_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_GLIN));
-	bws_in_list = g_slist_prepend(bws_in_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_GIN));
-	bws_in_list = g_slist_prepend(bws_in_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_GIN_UDP));
-	bws_in_list = g_slist_prepend(bws_in_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_IN));
-	bws_in_list = g_slist_prepend(bws_in_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_DHT_IN));
+	bws_in_list = pslist_prepend(bws_in_list,
+						uint_to_pointer(BSCHED_BWS_LOOPBACK_IN));
+	bws_in_list = pslist_prepend(bws_in_list,
+						uint_to_pointer(BSCHED_BWS_PRIVATE_IN));
+	bws_in_list = pslist_prepend(bws_in_list,
+						uint_to_pointer(BSCHED_BWS_GLIN));
+	bws_in_list = pslist_prepend(bws_in_list,
+						uint_to_pointer(BSCHED_BWS_GIN));
+	bws_in_list = pslist_prepend(bws_in_list,
+						uint_to_pointer(BSCHED_BWS_GIN_UDP));
+	bws_in_list = pslist_prepend(bws_in_list,
+						uint_to_pointer(BSCHED_BWS_IN));
+	bws_in_list = pslist_prepend(bws_in_list,
+						uint_to_pointer(BSCHED_BWS_DHT_IN));
 
-	bws_out_list = g_slist_prepend(bws_out_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_LOOPBACK_OUT));
-	bws_out_list = g_slist_prepend(bws_out_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_PRIVATE_OUT));
-	bws_out_list = g_slist_prepend(bws_out_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_GLOUT));
-	bws_out_list = g_slist_prepend(bws_out_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_GOUT));
-	bws_out_list = g_slist_prepend(bws_out_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_GOUT_UDP));
-	bws_out_list = g_slist_prepend(bws_out_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_OUT));
-	bws_out_list = g_slist_prepend(bws_out_list, 
-						GUINT_TO_POINTER(BSCHED_BWS_DHT_OUT));
+	bws_out_list = pslist_prepend(bws_out_list,
+						uint_to_pointer(BSCHED_BWS_LOOPBACK_OUT));
+	bws_out_list = pslist_prepend(bws_out_list,
+						uint_to_pointer(BSCHED_BWS_PRIVATE_OUT));
+	bws_out_list = pslist_prepend(bws_out_list,
+						uint_to_pointer(BSCHED_BWS_GLOUT));
+	bws_out_list = pslist_prepend(bws_out_list,
+						uint_to_pointer(BSCHED_BWS_GOUT));
+	bws_out_list = pslist_prepend(bws_out_list,
+						uint_to_pointer(BSCHED_BWS_GOUT_UDP));
+	bws_out_list = pslist_prepend(bws_out_list,
+						uint_to_pointer(BSCHED_BWS_OUT));
+	bws_out_list = pslist_prepend(bws_out_list,
+						uint_to_pointer(BSCHED_BWS_DHT_OUT));
 }
 
 /**
@@ -607,17 +609,17 @@ bsched_init(void)
 G_GNUC_COLD void
 bsched_close(void)
 {
-	GSList *iter;
+	pslist_t *iter;
 	uint i;
 
-	for (iter = bws_list; iter; iter = g_slist_next(iter)) {
-		bsched_bws_t bws = GPOINTER_TO_UINT(iter->data);
+	PSLIST_FOREACH(bws_list, iter) {
+		bsched_bws_t bws = pointer_to_uint(iter->data);
 		bsched_free(bsched_get(bws));
 	}
 
-	gm_slist_free_null(&bws_list);
-	gm_slist_free_null(&bws_out_list);
-	gm_slist_free_null(&bws_in_list);
+	pslist_free_null(&bws_list);
+	pslist_free_null(&bws_out_list);
+	pslist_free_null(&bws_in_list);
 
 	for (i = 0; i < NUM_BSCHED_BWS; i++) {
 		bws_set[i] = NULL;
@@ -771,10 +773,10 @@ bsched_enable_all(void)
 G_GNUC_COLD void
 bsched_shutdown(void)
 {
-	GSList *sl;
+	pslist_t *sl;
 
-	for (sl = bws_list; sl; sl = g_slist_next(sl)) {
-		bsched_bws_t bws = GPOINTER_TO_UINT(sl->data);
+	PSLIST_FOREACH(bws_list, sl) {
+		bsched_bws_t bws = pointer_to_uint(sl->data);
 		bsched_disable(bws);
 	}
 }
@@ -931,10 +933,10 @@ bio_remove_callback(bio_source_t *bio)
 static void
 bsched_no_more_bandwidth(bsched_t *bs)
 {
-	GList *iter;
+	plist_t *iter;
 
 	bsched_check(bs);
-	for (iter = bs->sources; iter; iter = g_list_next(iter)) {
+	PLIST_FOREACH(bs->sources, iter) {
 		bio_source_t *bio = iter->data;
 
 		bio_check(bio);
@@ -952,10 +954,10 @@ bsched_no_more_bandwidth(bsched_t *bs)
 static void
 bsched_clear_active(bsched_t *bs)
 {
-	GList *iter;
+	plist_t *iter;
 
 	bsched_check(bs);
-	for (iter = bs->sources; iter; iter = g_list_next(iter)) {
+	PLIST_FOREACH(bs->sources, iter) {
 		bio_source_t *bio = iter->data;
 
 		bio_check(bio);
@@ -973,9 +975,8 @@ bsched_clear_active(bsched_t *bs)
 static void
 bsched_begin_timeslice(bsched_t *bs)
 {
-	GList *iter;
-	GList *last = NULL;
-	GSList *trigger = NULL;
+	plist_t *iter, *last = NULL;
+	pslist_t *trigger = NULL;
 	double norm_factor;
 	int count;
 	unsigned bw_max;
@@ -1004,8 +1005,7 @@ bsched_begin_timeslice(bsched_t *bs)
 		if (bs->bw_stolen != 0) {
 			if (GNET_PROPERTY(bsched_debug)) {
 				g_debug("BSCHED %s: \"%s\" ignoring stolen %d byte%s",
-					G_STRFUNC, bs->name, bs->bw_stolen,
-					1 == bs->bw_stolen ? "" : "s");
+					G_STRFUNC, bs->name, bs->bw_stolen, plural(bs->bw_stolen));
 			}
 			bs->bw_stolen = 0;
 		}
@@ -1013,7 +1013,7 @@ bsched_begin_timeslice(bsched_t *bs)
 			if (GNET_PROPERTY(bsched_debug)) {
 				g_debug("BSCHED %s: \"%s\" ignoring %d byte%s of capped b/w",
 					G_STRFUNC, bs->name, bs->bw_last_capped,
-					1 == bs->bw_last_capped ? "" : "s");
+					plural(bs->bw_last_capped));
 			}
 			bs->bw_last_capped = 0;
 		}
@@ -1022,8 +1022,9 @@ bsched_begin_timeslice(bsched_t *bs)
 	norm_factor = 1000.0 / bs->period;
 	bs->io_favours = 0;
 	bw_max = bs->bw_max;
+	count = 0;
 
-	for (count = 0, iter = bs->sources; iter; iter = g_list_next(iter)) {
+	PLIST_FOREACH(bs->sources, iter) {
 		bio_source_t *bio = iter->data;
 		uint32 actual;
 
@@ -1036,7 +1037,7 @@ bsched_begin_timeslice(bsched_t *bs)
 
 		if (bio->io_tag == 0 && bio->io_callback) {
 			if (bio->flags & BIO_F_PASSIVE)
-				trigger = g_slist_prepend(trigger, bio);
+				trigger = pslist_prepend(trigger, bio);
 			else
 				bio_enable(bio);
 		}
@@ -1091,8 +1092,8 @@ bsched_begin_timeslice(bsched_t *bs)
 		g_assert(bs->sources != NULL);
 		bio = bs->sources->data;
 		bio_check(bio);
-		bs->sources = g_list_remove(bs->sources, bio);
-		bs->sources = gm_list_insert_after(bs->sources, last, bio);
+		bs->sources = plist_remove(bs->sources, bio);
+		bs->sources = plist_insert_after(bs->sources, last, bio);
 	}
 
 	bs->flags &= ~(BS_F_NOBW|BS_F_FROZEN_SLOT|BS_F_CHANGED_BW|BS_F_CLEARED);
@@ -1165,7 +1166,7 @@ bsched_begin_timeslice(bsched_t *bs)
 	while (trigger != NULL) {
 		bio_source_t *bio = trigger->data;
 
-		trigger = g_slist_remove(trigger, bio);
+		trigger = pslist_remove(trigger, bio);
 		bio_trigger(bio);
 	}
 }
@@ -1179,7 +1180,7 @@ bsched_bio_add(bsched_t *bs, bio_source_t *bio)
 	bsched_check(bs);
 	bio_check(bio);
 
-	bs->sources = g_list_append(bs->sources, bio);
+	bs->sources = plist_append(bs->sources, bio);
 	bs->count++;
 
 	bs->bw_slot = (bs->bw_max + bs->bw_stolen) / bs->count;
@@ -1204,7 +1205,7 @@ bsched_bio_remove(bsched_bws_t bws, bio_source_t *bio)
 	bs = bsched_get(bws);
 	bio_check(bio);
 
-	bs->sources = g_list_remove(bs->sources, bio);
+	bs->sources = plist_remove(bs->sources, bio);
 	bs->count--;
 
 	if (bs->count)
@@ -2145,7 +2146,7 @@ bio_sendfile(sendfile_ctx_t *ctx, bio_source_t *bio,
 		g_warning("FIXED SENDFILE returned offset: "
 			"was set to %s instead of %s (%zu byte%s written)",
 			uint64_to_string(*offset), uint64_to_string2(start + r),
-			r, r == 1 ? "" : "s");
+			r, plural(r));
 		*offset = start + r;
 	} else if ((ssize_t) -1 == r) {
 		*offset = start;	/* Paranoid: in case sendfile() touched it */
@@ -2501,7 +2502,7 @@ bs_socket(enum socket_direction dir, enum socket_type type)
 		break;
 	}
 	if (BSCHED_BWS_INVALID == bws) {
-		g_warning("bs_socket: unhandled socket type %d", type);
+		g_warning("%s(): unhandled socket type %d", G_STRFUNC, type);
 		bws = SOCK_CONN_OUTGOING == dir ? BSCHED_BWS_OUT : BSCHED_BWS_IN;
 	}
 	return bsched_get(bws);
@@ -2678,7 +2679,7 @@ bws_can_connect(enum socket_type type)
 static void
 bsched_heartbeat(bsched_t *bs, tm_t *tv)
 {
-	GList *iter;
+	plist_t *iter;
 	int delay;
 	int overused;
 	int theoric;
@@ -2768,6 +2769,8 @@ bsched_heartbeat(bsched_t *bs, tm_t *tv)
 
 	overused -= bs->bw_stolen;		/* Correct for computations below */
 
+	entropy_harvest_single(VARLEN(overused));
+
 	bs->bw_max = (int) (bs->bw_per_second / 1000.0 * bs->period_ema);
 
 	/*
@@ -2822,7 +2825,7 @@ bsched_heartbeat(bsched_t *bs, tm_t *tv)
 
 	last_used = 0;
 
-	for (iter = bs->sources; iter; iter = g_list_next(iter)) {
+	PLIST_FOREACH(bs->sources, iter) {
 		bio_source_t *bio = iter->data;
 
 		bio_check(bio);
@@ -2849,7 +2852,7 @@ bsched_heartbeat(bsched_t *bs, tm_t *tv)
 			bs->count ? bs->bw_max / bs->count : 0,
 			bs->count ? (bs->bw_max + bs->bw_capped) / bs->count : 0,
 			bs->bw_per_second, bs->count,
-			bs->count == 1 ? "" : "s", bs->bw_actual * 1000.0 / delay);
+			plural(bs->count), bs->bw_actual * 1000.0 / delay);
 	}
 
 	/*
@@ -2869,8 +2872,8 @@ new_timeslice:
 static void
 bsched_stealbeat(bsched_t *bs)
 {
-	GSList *l;
-	GSList *all_used = NULL;		/* List of bsched_t that used all b/w */
+	pslist_t *l;
+	pslist_t *all_used = NULL;		/* List of bsched_t that used all b/w */
 	int all_used_count = 0;			/* Amount of bsched_t that used all b/w */
 	int all_favour_count = 0;		/* I/O sources wanting favours */
 	uint all_bw_count = 0;			/* Sum of configured bandwidth */
@@ -2926,9 +2929,9 @@ bsched_stealbeat(bsched_t *bs)
 
 	if (bs->flags & BS_F_WRITE) {
 		int half_contribution = bs->count ? bs->bw_max / (2 * bs->count) : 0;
-		GList *bl;
+		plist_t *bl;
 
-		for (bl = bs->sources; bl && underused > 0; bl = g_list_next(bl)) {
+		for (bl = bs->sources; bl && underused > 0; bl = plist_next(bl)) {
 			bio_source_t *bio = (bio_source_t *) bl->data;
 
 			if (bio->io_callback != NULL && !(bio->flags & BIO_F_USED))
@@ -2937,6 +2940,8 @@ bsched_stealbeat(bsched_t *bs)
 	}
 #endif
 
+	entropy_harvest_single(VARLEN(underused));
+
 	if (underused <= 0)				/* Nothing to redistribute */
 		return;
 
@@ -2944,14 +2949,14 @@ bsched_stealbeat(bsched_t *bs)
 	 * Determine who used up all its bandwidth among our stealers.
 	 */
 
-	for (l = bs->stealers; l; l = g_slist_next(l)) {
+	PSLIST_FOREACH(bs->stealers, l) {
 		bsched_t *xbs = l->data;
 
 		steal_count++;
 		all_favour_count += xbs->io_favours;
 
 		if (xbs->bw_last_period >= xbs->bw_max) {
-			all_used = g_slist_prepend(all_used, xbs);
+			all_used = pslist_prepend(all_used, xbs);
 			all_used_count++;
 			all_bw_count += xbs->bw_max;
 		}
@@ -2992,7 +2997,7 @@ bsched_stealbeat(bsched_t *bs)
 	 */
 
 	if (all_favour_count != 0) {
-		for (l = bs->stealers; l; l = g_slist_next(l)) {
+		PSLIST_FOREACH(bs->stealers, l) {
 			bsched_t *xbs = l->data;
 			double amount;
 
@@ -3011,10 +3016,10 @@ bsched_stealbeat(bsched_t *bs)
 				g_debug("BSCHED %s: \"%s\" giving %d bytes to \"%s\" "
 					"(%d favour%s)",
 					G_STRFUNC, bs->name, (int) amount, xbs->name,
-					xbs->io_favours, 1 == xbs->io_favours ? "" : "s");
+					xbs->io_favours, plural(xbs->io_favours));
 		}
 	} else if (all_used_count == 0) {
-		for (l = bs->stealers; l; l = g_slist_next(l)) {
+		PSLIST_FOREACH(bs->stealers, l) {
 			bsched_t *xbs = l->data;
 			xbs->bw_stolen += underused / steal_count;
 
@@ -3023,7 +3028,7 @@ bsched_stealbeat(bsched_t *bs)
 					G_STRFUNC, bs->name, underused / steal_count, xbs->name);
 		}
 	} else {
-		for (l = all_used; l; l = g_slist_next(l)) {
+		PSLIST_FOREACH(all_used, l) {
 			bsched_t *xbs = l->data;
 			double amount;
 
@@ -3044,7 +3049,7 @@ bsched_stealbeat(bsched_t *bs)
 	}
 
 done:
-	g_slist_free(all_used);
+	pslist_free(all_used);
 }
 
 /**
@@ -3054,7 +3059,7 @@ void
 bsched_timer(void)
 {
 	tm_t tv;
-	GSList *l;
+	pslist_t *l;
 	int out_used = 0;
 	int in_used = 0;
 	bool read_data = FALSE;
@@ -3065,8 +3070,8 @@ bsched_timer(void)
 	 * First pass: compute bandwidth used.
 	 */
 
-	for (l = bws_list; l; l = g_slist_next(l)) {
-		bsched_bws_t bws = GPOINTER_TO_UINT(l->data);
+	PSLIST_FOREACH(bws_list, l) {
+		bsched_bws_t bws = pointer_to_uint(l->data);
 		bsched_heartbeat(bsched_get(bws), &tv);
 	}
 
@@ -3075,8 +3080,8 @@ bsched_timer(void)
 	 * have not used up all their quota.
 	 */
 
-	for (l = bws_list; l; l = g_slist_next(l)) {
-		bsched_bws_t bws = GPOINTER_TO_UINT(l->data);
+	PSLIST_FOREACH(bws_list, l) {
+		bsched_bws_t bws = pointer_to_uint(l->data);
 		bsched_stealbeat(bsched_get(bws));
 	}
 
@@ -3084,8 +3089,8 @@ bsched_timer(void)
 	 * Third pass: begin new timeslice.
 	 */
 
-	for (l = bws_list; l; l = g_slist_next(l)) {
-		bsched_bws_t bws = GPOINTER_TO_UINT(l->data);
+	PSLIST_FOREACH(bws_list, l) {
+		bsched_bws_t bws = pointer_to_uint(l->data);
 		bsched_begin_timeslice(bsched_get(bws));
 	}
 
@@ -3093,8 +3098,8 @@ bsched_timer(void)
 	 * Fourth pass: update the average bandwidth used.
 	 */
 
-	for (l = bws_out_list; l; l = g_slist_next(l)) {
-		bsched_bws_t bws = GPOINTER_TO_UINT(l->data);
+	PSLIST_FOREACH(bws_out_list, l) {
+		bsched_bws_t bws = pointer_to_uint(l->data);
 		bsched_t *bs = bsched_get(bws);
 		out_used += (int) (bs->bw_last_period * 1000.0 / bs->period_ema);
 	}
@@ -3104,8 +3109,8 @@ bsched_timer(void)
 	if (GNET_PROPERTY(bsched_debug) > 3)
 		g_debug("BSCHED outgoing b/w EMA = %d bytes/s", bws_out_ema);
 
-	for (l = bws_in_list; l; l = g_slist_next(l)) {
-		bsched_bws_t bws = GPOINTER_TO_UINT(l->data);
+	PSLIST_FOREACH(bws_in_list, l) {
+		bsched_bws_t bws = pointer_to_uint(l->data);
 		bsched_t *bs = bsched_get(bws);
 
 		in_used += (int) (bs->bw_last_period * 1000.0 / bs->period_ema);
@@ -3113,6 +3118,7 @@ bsched_timer(void)
 		if (bs->flags & BS_F_DATA_READ) {
 			read_data = TRUE;
 			bs->flags &= ~BS_F_DATA_READ;
+			entropy_harvest_single(VARLEN(in_used));
 		}
 	}
 
