@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2001-2003, Richard Eckart
+ * Copyright (c) 2008-2014, Raphael Manfredi
  *
  *----------------------------------------------------------------------
  * This file is part of gtk-gnutella.
@@ -25,10 +26,12 @@
  * @ingroup core
  * @file
  *
- * Needs brief description here.
+ * Collection of Gnutella / DHT statistics.
  *
  * @author Richard Eckart
  * @date 2001-2003
+ * @author Raphael Manfredi
+ * @date 2008-2014
  */
 
 #include "common.h"
@@ -36,14 +39,19 @@
 #include "gnet_stats.h"
 #include "gmsg.h"
 
+#include "g2/msg.h"
+
 #include "if/dht/kademlia.h"
 #include "if/gnet_property_priv.h"
 
-#include "lib/crc.h"
+#include "lib/entropy.h"
 #include "lib/event.h"
-#include "lib/gnet_host.h"
 #include "lib/random.h"
+#include "lib/sha1.h"
+#include "lib/spinlock.h"
+#include "lib/thread.h"
 #include "lib/tm.h"
+
 #include "lib/override.h"		/* Must be the last header included */
 
 static uint8 stats_lut[256];
@@ -52,366 +60,25 @@ static gnet_stats_t gnet_stats;
 static gnet_stats_t gnet_tcp_stats;
 static gnet_stats_t gnet_udp_stats;
 
-static uint32 gnet_stats_crc32;
+/*
+ * Thread-safe locks.
+ *
+ * The general stats accounting code is protected because there is no guarantee
+ * the routines updating these stats will always be called from the main thread.
+ *
+ * However, the routines updating the traffic statistics are NOT protected
+ * because they are always called from the main thread, the one where the
+ * I/O event loop is installed.  An assertion verifies this assumption.
+ */
+static spinlock_t gnet_stats_slk = SPINLOCK_INIT;
+
+#define GNET_STATS_LOCK		spinlock_hidden(&gnet_stats_slk)
+#define GNET_STATS_UNLOCK	spinunlock_hidden(&gnet_stats_slk)
 
 /***
  *** Public functions
  ***/
 
-const char *
-gnet_stats_drop_reason_to_string(msg_drop_reason_t reason)
-{
-	static const char * const msg_drop_reasons[] = {
-		N_("Bad size"),						 /**< MSG_DROP_BAD_SIZE */
-		N_("Too small"),					 /**< MSG_DROP_TOO_SMALL */
-		N_("Too large"),					 /**< MSG_DROP_TOO_LARGE */
-		N_("Way too large"),				 /**< MSG_DROP_WAY_TOO_LARGE */
-		N_("Too old"),					 	 /**< MSG_DROP_TOO_OLD */
-		N_("Unknown message type"),			 /**< MSG_DROP_UNKNOWN_TYPE */
-		N_("Unexpected message"),			 /**< MSG_DROP_UNEXPECTED */
-		N_("Message sent with TTL = 0"),	 /**< MSG_DROP_TTL0 */
-		N_("Improper hops/ttl combination"), /**< MSG_DROP_IMPROPER_HOPS_TTL */
-		N_("Max TTL exceeded"),				 /**< MSG_DROP_MAX_TTL_EXCEEDED */
-		N_("Message throttle"),				 /**< MSG_DROP_THROTTLE */
-		N_("Message matched limits"),		 /**< MSG_DROP_LIMIT */
-		N_("Transient node"),				 /**< MSG_DROP_TRANSIENT */
-		N_("Unusable Pong"),				 /**< MSG_DROP_PONG_UNUSABLE */
-		N_("Hard TTL limit reached"),		 /**< MSG_DROP_HARD_TTL_LIMIT */
-		N_("Max hop count reached"),		 /**< MSG_DROP_MAX_HOP_COUNT */
-		N_("Route lost"),					 /**< MSG_DROP_ROUTE_LOST */
-		N_("No route"),						 /**< MSG_DROP_NO_ROUTE */
-		N_("Duplicate message"),			 /**< MSG_DROP_DUPLICATE */
-		N_("Message to banned GUID"),		 /**< MSG_DROP_TO_BANNED */
-		N_("Message from banned GUID"),		 /**< MSG_DROP_FROM_BANNED */
-		N_("Node shutting down"),			 /**< MSG_DROP_SHUTDOWN */
-		N_("TX flow control"),				 /**< MSG_DROP_FLOW_CONTROL */
-		N_("Query text had no trailing NUL"),/**< MSG_DROP_QUERY_NO_NUL */
-		N_("Query text too short"),			 /**< MSG_DROP_QUERY_TOO_SHORT */
-		N_("Query had unnecessary overhead"),/**< MSG_DROP_QUERY_OVERHEAD */
-		N_("Query had bad URN"),			 /**< MSG_DROP_BAD_URN */
-		N_("Message with malformed SHA1"),	 /**< MSG_DROP_MALFORMED_SHA1 */
-		N_("Message with malformed UTF-8"),	 /**< MSG_DROP_MALFORMED_UTF_8 */
-		N_("Malformed Query Hit"),			 /**< MSG_DROP_BAD_RESULT */
-		N_("Bad return address"),			 /**< MSG_DROP_BAD_RETURN_ADDRESS */
-		N_("Hostile IP address"),			 /**< MSG_DROP_HOSTILE_IP */
-		N_("Bogus result from Morpheus"),	 /**< MSG_DROP_MORPHEUS_BOGUS */
-		N_("Spam"),							 /**< MSG_DROP_SPAM */
-		N_("Evil filename"),				 /**< MSG_DROP_EVIL */
-		N_("Improper media type"),			 /**< MSG_DROP_MEDIA */
-		N_("Payload inflating error"),		 /**< MSG_DROP_INFLATE_ERROR */
-		N_("Unknown header flags present"),/**< MSG_DROP_UNKNOWN_HEADER_FLAGS */
-		N_("Own search results"),			 /**< MSG_DROP_OWN_RESULT */
-		N_("Own queries"),			 		 /**< MSG_DROP_OWN_QUERY */
-		N_("Ancient query format"),			 /**< MSG_DROP_ANCIENT_QUERY */
-		N_("Blank Servent ID"),				 /**< MSG_DROP_BLANK_SERVENT_ID */
-		N_("GUESS Query missing token"), /**< MSG_DROP_GUESS_MISSING_TOKEN */
-		N_("GUESS Invalid query token"), /**< MSG_DROP_GUESS_INVALID_TOKEN */
-		N_("DHT Invalid security token"),	 /**< MSG_DROP_DHT_INVALID_TOKEN */
-		N_("DHT Too many STORE requests"),	 /**< MSG_DROP_DHT_TOO_MANY_STORE */
-		N_("DHT Malformed message"),		 /**< MSG_DROP_DHT_UNPARSEABLE */
-	};
-
-	STATIC_ASSERT(G_N_ELEMENTS(msg_drop_reasons) == MSG_DROP_REASON_COUNT);
-	g_return_val_if_fail(UNSIGNED(reason) < G_N_ELEMENTS(msg_drop_reasons),
-		NULL);
-	return msg_drop_reasons[reason];
-}
-
-const char *
-gnet_stats_general_to_string(gnr_stats_t type)
-{
-	/* Do NOT translate any of these strings */
-
-	static const char * const type_string[] = {
-		"routing_errors",
-		"routing_table_chunks",
-		"routing_table_capacity",
-		"routing_table_count",
-		"routing_transient_avoided",
-		"dups_with_higher_ttl",
-		"spam_sha1_hits",
-		"spam_name_hits",
-		"spam_fake_hits",
-		"spam_dup_hits",
-		"spam_caught_hostile_ip",
-		"spam_caught_hostile_held",
-		"spam_ip_held",
-		"local_searches",
-		"local_hits",
-		"local_partial_hits",
-		"local_whats_new_hits",
-		"local_query_hits",
-		"oob_proxied_query_hits",
-		"oob_queries",
-		"oob_queries_stripped",
-		"oob_queries_ignored",
-		"query_oob_proxied_dups",
-		"oob_hits_for_proxied_queries",
-		"oob_hits_with_alien_ip",
-		"oob_hits_ignored_on_spammer_hit",
-		"oob_hits_ignored_on_unsecure_hit",
-		"unclaimed_oob_hits",
-		"partially_claimed_oob_hits",
-		"spurious_oob_hit_claim",
-		"unrequested_oob_hits",
-		"query_hit_for_untracked_query",
-		"query_tracked_muids",
-		"query_compact_count",
-		"query_compact_size",
-		"query_utf8",
-		"query_sha1",
-		"query_whats_new",
-		"query_guess",
-		"query_guess_02",
-		"guess_link_cache",
-		"guess_cached_query_keys_held",
-		"guess_cached_02_hosts_held",
-		"guess_local_queries",
-		"guess_local_running",
-		"guess_local_query_hits",
-		"guess_hosts_queried",
-		"guess_hosts_acknowledged",
-		"broadcasted_pushes",
-		"push_proxy_udp_relayed",
-		"push_proxy_tcp_relayed",
-		"push_proxy_tcp_fw2fw",
-		"push_proxy_broadcasted",
-		"push_proxy_route_not_proxied",
-		"push_proxy_failed",
-		"push_relayed_via_local_route",
-		"push_relayed_via_table_route",
-		"local_dyn_queries",
-		"leaf_dyn_queries",
-		"oob_proxied_queries",
-		"dyn_queries_completed_full",
-		"dyn_queries_completed_partial",
-		"dyn_queries_completed_zero",
-		"dyn_queries_linger_extra",
-		"dyn_queries_linger_results",
-		"dyn_queries_linger_completed",
-		"gtkg_total_queries",
-		"gtkg_requeries",
-		"queries_with_ggep_h",
-		"queries_with_sr_udp",
-		"giv_callbacks",
-		"giv_discarded",
-		"queue_callbacks",
-		"queue_discarded",
-		"udp_read_ahead_count_sum",
-		"udp_read_ahead_bytes_sum",
-		"udp_read_ahead_old_sum",
-		"udp_read_ahead_count_max",
-		"udp_read_ahead_bytes_max",
-		"udp_read_ahead_delay_max",
-		"udp_fw2fw_pushes",
-		"udp_fw2fw_pushes_to_self",
-		"udp_fw2fw_pushes_patched",
-		"udp_uhc_pings",
-		"udp_uhc_pongs",
-		"udp_bogus_source_ip",
-		"udp_rx_truncated",
-		"udp_alien_message",
-		"udp_unprocessed_message",
-		"udp_tx_compressed",
-		"udp_rx_compressed",
-		"udp_larger_hence_not_compressed",
-		"udp_ambiguous",
-		"udp_ambiguous_deeper_inspection",
-		"udp_ambiguous_as_semi_reliable",
-		"udp_sr_tx_messages_given",
-		"udp_sr_tx_messages_deflated",
-		"udp_sr_tx_messages_unsent",
-		"udp_sr_tx_messages_banned",
-		"udp_sr_tx_messages_clogging",
-		"udp_sr_tx_reliable_messages_given",
-		"udp_sr_tx_reliable_messages_sent",
-		"udp_sr_tx_reliable_messages_unsent",
-		"udp_sr_tx_fragments_sent",
-		"udp_sr_tx_fragments_resent",
-		"udp_sr_tx_fragments_sending_avoided",
-		"udp_sr_tx_fragments_oversent",
-		"udp_sr_tx_total_acks_received",
-		"udp_sr_tx_cumulative_acks_received",
-		"udp_sr_tx_extended_acks_received",
-		"udp_sr_tx_spurious_acks_received",
-		"udp_sr_tx_invalid_acks_received",
-		"udp_sr_tx_ears_sent",
-		"udp_sr_tx_ears_oversent",
-		"udp_sr_tx_ear_nacks_received",
-		"udp_sr_tx_ear_followed_by_acks",
-		"udp_sr_rx_fragments_received",
-		"udp_sr_rx_fragments_duplicate",
-		"udp_sr_rx_fragments_unreliable",
-		"udp_sr_rx_fragments_dropped",
-		"udp_sr_rx_fragments_lingering",
-		"udp_sr_rx_messages_expired",
-		"udp_sr_rx_messages_received",
-		"udp_sr_rx_messages_inflated",
-		"udp_sr_rx_messages_inflation_error",
-		"udp_sr_rx_messages_unreliable",
-		"udp_sr_rx_messages_empty",
-		"udp_sr_rx_total_acks_sent",
-		"udp_sr_rx_cumulative_acks_sent",
-		"udp_sr_rx_extended_acks_sent",
-		"udp_sr_rx_avoided_acks",
-		"udp_sr_rx_ears_received",
-		"udp_sr_rx_ears_for_unknown_message",
-		"udp_sr_rx_ears_for_lingering_message",
-		"udp_sr_rx_from_hostile_ip",
-		"consolidated_servers",
-		"dup_downloads_in_consolidation",
-		"discovered_server_guid",
-		"changed_server_guid",
-		"guid_collisions",
-		"own_guid_collisions",
-		"banned_guid_held",
-		"received_known_fw_node_info",
-		"revitalized_push_routes",
-		"collected_push_proxies",
-		"attempted_resource_switching",
-		"attempted_resource_switching_after_error",
-		"successful_resource_switching",
-		"successful_plain_resource_switching",
-		"successful_resource_switching_after_error",
-		"queued_after_switching",
-		"sunk_data",
-		"ignored_data",
-		"ignoring_after_mismatch",
-		"ignoring_to_preserve_connection",
-		"ignoring_during_aggressive_swarming",
-		"ignoring_refused",
-		"client_resource_switching",
-		"client_plain_resource_switching",
-		"client_followup_after_error",
-		"parq_slot_resource_switching",
-		"parq_retry_after_violation",
-		"parq_retry_after_kick_out",
-		"parq_slot_limit_overrides",
-		"parq_quick_slots_granted",
-		"parq_queue_sending_attempts",
-		"parq_queue_sent",
-		"parq_queue_follow_ups",
-		"sha1_verifications",
-		"tth_verifications",
-		"bitzi_tickets_held",
-		"qhit_seeding_of_orphan",
-		"upload_seeding_of_orphan",
-		"rudp_tx_bytes",
-		"rudp_rx_bytes",
-		"dht_estimated_size",
-		"dht_estimated_size_stderr",
-		"dht_kball_theoretical",
-		"dht_kball_furthest",
-		"dht_kball_closest",
-		"dht_routing_buckets",
-		"dht_routing_leaves",
-		"dht_routing_max_depth",
-		"dht_routing_good_nodes",
-		"dht_routing_stale_nodes",
-		"dht_routing_pending_nodes",
-		"dht_routing_evicted_nodes",
-		"dht_routing_evicted_firewalled_nodes",
-		"dht_routing_evicted_quota_nodes",
-		"dht_routing_promoted_pending_nodes",
-		"dht_routing_pinged_promoted_nodes",
-		"dht_routing_rejected_node_bucket_quota",
-		"dht_routing_rejected_node_global_quota",
-		"dht_completed_bucket_refresh",
-		"dht_forced_bucket_refresh",
-		"dht_forced_bucket_merge",
-		"dht_denied_unsplitable_bucket_refresh",
-		"dht_bucket_alive_check",
-		"dht_alive_pings_to_good_nodes",
-		"dht_alive_pings_to_stale_nodes",
-		"dht_alive_pings_to_shutdowning_nodes",
-		"dht_alive_pings_avoided",
-		"dht_alive_pings_skipped",
-		"dht_revitalized_stale_nodes",
-		"dht_rejected_value_on_quota",
-		"dht_rejected_value_on_creator",
-		"dht_lookup_rejected_node_on_net_quota",
-		"dht_lookup_rejected_node_on_proximity",
-		"dht_lookup_rejected_node_on_divergence",
-		"dht_lookup_fixed_node_contact",
-		"dht_keys_held",
-		"dht_cached_keys_held",
-		"dht_values_held",
-		"dht_cached_kuid_targets_held",
-		"dht_cached_roots_held",
-		"dht_cached_roots_exact_hits",
-		"dht_cached_roots_approximate_hits",
-		"dht_cached_roots_misses",
-		"dht_cached_roots_kball_lookups",
-		"dht_cached_roots_contact_refreshed",
-		"dht_cached_tokens_held",
-		"dht_cached_tokens_hits",
-		"dht_stable_nodes_held",
-		"dht_fetch_local_hits",
-		"dht_fetch_local_cached_hits",
-		"dht_returned_expanded_values",
-		"dht_returned_secondary_keys",
-		"dht_claimed_secondary_keys",
-		"dht_returned_expanded_cached_values",
-		"dht_returned_cached_secondary_keys",
-		"dht_claimed_cached_secondary_keys",
-		"dht_published",
-		"dht_removed",
-		"dht_stale_replication",
-		"dht_replication",
-		"dht_republish",
-		"dht_secondary_key_fetch",
-		"dht_dup_values",
-		"dht_kuid_collisions",
-		"dht_own_kuid_collisions",
-		"dht_caching_attempts",
-		"dht_caching_successful",
-		"dht_caching_partially_successful",
-		"dht_key_offloading_checks",
-		"dht_keys_selected_for_offloading",
-		"dht_key_offloading_attempts",
-		"dht_key_offloading_successful",
-		"dht_key_offloading_partially_successful",
-		"dht_values_offloaded",
-		"dht_msg_received",
-		"dht_msg_matching_contact_address",
-		"dht_msg_fixed_contact_address",
-		"dht_msg_from_hostile_address",
-		"dht_msg_from_hostile_contact_address",
-		"dht_rpc_msg_prepared",
-		"dht_rpc_msg_cancelled",
-		"dht_rpc_timed_out",
-		"dht_rpc_replies_received",
-		"dht_rpc_replies_fixed_contact",
-		"dht_rpc_late_replies_received",
-		"dht_rpc_kuid_reply_mismatch",
-		"dht_rpc_recent_nodes_held",
-		"dht_node_verifications",
-		"dht_publishing_attempts",
-		"dht_publishing_successful",
-		"dht_publishing_partially_successful",
-		"dht_publishing_satisfactory",
-		"dht_republished_late",
-		"dht_publishing_to_self",
-		"dht_publishing_bg_attempts",
-		"dht_publishing_bg_improvements",
-		"dht_publishing_bg_successful",
-		"dht_sha1_data_type_collisions",
-		"dht_passively_protected_lookup_path",
-		"dht_actively_protected_lookup_path",
-		"dht_alt_loc_lookups",
-		"dht_push_proxy_lookups",
-		"dht_successful_alt_loc_lookups",
-		"dht_successful_push_proxy_lookups",
-		"dht_successful_node_push_entry_lookups",
-		"dht_seeding_of_orphan",
-	};
-
-	STATIC_ASSERT(G_N_ELEMENTS(type_string) == GNR_TYPE_COUNT);
-	g_return_val_if_fail(UNSIGNED(type) < G_N_ELEMENTS(type_string),
-		NULL);
-	return type_string[type];
-}
-
 G_GNUC_COLD void
 gnet_stats_init(void)
 {
@@ -420,6 +87,8 @@ gnet_stats_init(void)
 	/* Guarantees that our little hack below can succeed */
 	STATIC_ASSERT(
 		UNSIGNED(KDA_MSG_MAX_ID + MSG_DHT_BASE) < G_N_ELEMENTS(stats_lut));
+	STATIC_ASSERT(
+		UNSIGNED(MSG_G2_BASE + G2_MSG_MAX) < GTA_MSG_QRP);
 
 	for (i = 0; i < G_N_ELEMENTS(stats_lut); i++) {
 		uchar m = MSG_UNKNOWN;
@@ -432,6 +101,10 @@ gnet_stats_init(void)
 		 * by Gnutella to stuff the DHT messages there.  And 0xd0 starts
 		 * with a 'D', so it's not a total hack.
 		 *		--RAM, 2010-11-01.
+		 *
+		 * We play the same trick for G2 messages, only we insert them
+		 * in the 0x05 .. 0x2f space, which is unused by Gnutella.
+		 *		--RAM, 2014-01-07.
 		 */
 
 		if (i > MSG_DHT_BASE) {
@@ -449,6 +122,31 @@ gnet_stats_init(void)
 				/* deprecated, not supported */
 				break;
 			}
+		} else if (i > MSG_G2_BASE && i < GTA_MSG_QRP) {
+			switch ((enum g2_msg) (i - MSG_G2_BASE)) {
+			case G2_MSG_CRAWLR:			m = MSG_G2_CRAWLR; break;
+			case G2_MSG_HAW:			m = MSG_G2_HAW; break;
+			case G2_MSG_KHL:			m = MSG_G2_KHL; break;
+			case G2_MSG_KHLR:			m = MSG_G2_KHLR; break;
+			case G2_MSG_KHLA:			m = MSG_G2_KHLA; break;
+			case G2_MSG_LNI:			m = MSG_G2_LNI; break;
+			case G2_MSG_PI:				m = MSG_G2_PI; break;
+			case G2_MSG_PO:				m = MSG_G2_PO; break;
+			case G2_MSG_PUSH:			m = MSG_G2_PUSH; break;
+			case G2_MSG_QKA:			m = MSG_G2_QKA; break;
+			case G2_MSG_QKR:			m = MSG_G2_QKR; break;
+			case G2_MSG_Q2:				m = MSG_G2_Q2; break;
+			case G2_MSG_QA:				m = MSG_G2_QA; break;
+			case G2_MSG_QH2:			m = MSG_G2_QH2; break;
+			case G2_MSG_QHT:			m = MSG_G2_QHT; break;
+			case G2_MSG_UPROC:			m = MSG_G2_UPROC; break;
+			case G2_MSG_UPROD:			m = MSG_G2_UPROD; break;
+			case G2_MSG_MAX:
+				break;
+			case G2_MSG_CRAWLA:
+				/* This message is skipped, since we don't expect it */
+				g_assert_not_reached();
+			}
 		} else {
 			switch ((enum gta_msg) i) {
 			case GTA_MSG_INIT:           m = MSG_INIT; break;
@@ -463,30 +161,62 @@ gnet_stats_init(void)
 			case GTA_MSG_HSEP_DATA:		 m = MSG_HSEP; break;
 			case GTA_MSG_BYE:      		 m = MSG_BYE; break;
 			case GTA_MSG_DHT:            m = MSG_DHT; break;
+			case GTA_MSG_G2_SEARCH:	/* Not a real message */
 				break;
 			}
 		}
 		stats_lut[i] = m;
 	}
 
+	/* gnet_stats_count_received_payload() relies on this for G2 messages */
+	g_assert(MSG_UNKNOWN == stats_lut[G_N_ELEMENTS(stats_lut) - 1]);
+
 #undef CASE
 		
     ZERO(&gnet_stats);
     ZERO(&gnet_udp_stats);
-
-	gnet_stats_crc32 = random_u32();
 }
 
 /**
- * @return current CRC32 and re-initialize a new random one.
+ * Generate a SHA1 digest of the supplied statistics.
  */
-uint32
-gnet_stats_crc_reset(void)
+static void
+gnet_stats_digest(sha1_t *digest, const gnet_stats_t *stats)
 {
-	uint32 crc = gnet_stats_crc32;
+	SHA1_COMPUTE(*stats, digest);
+}
 
-	gnet_stats_crc32 = random_u32();
-	return crc;
+/**
+ * Generate a SHA1 digest of the current TCP statistics.
+ *
+ * This is meant for dynamic entropy collection.
+ */
+void
+gnet_stats_tcp_digest(sha1_t *digest)
+{
+	gnet_stats_digest(digest, &gnet_tcp_stats);
+}
+
+/**
+ * Generate a SHA1 digest of the current UDP statistics.
+ *
+ * This is meant for dynamic entropy collection.
+ */
+void
+gnet_stats_udp_digest(sha1_t *digest)
+{
+	gnet_stats_digest(digest, &gnet_udp_stats);
+}
+
+/**
+ * Generate a SHA1 digest of the current general statistics.
+ *
+ * This is meant for dynamic entropy collection.
+ */
+void
+gnet_stats_general_digest(sha1_t *digest)
+{
+	SHA1_COMPUTE(gnet_stats.general, digest);
 }
 
 /**
@@ -495,16 +225,36 @@ gnet_stats_crc_reset(void)
 static void
 gnet_stats_randomness(const gnutella_node_t *n, uint8 type, uint32 val)
 {
-	tm_t now;
-	gnet_host_t host;
+	entropy_harvest_small(
+		VARLEN(n->addr), VARLEN(n->port), VARLEN(type), VARLEN(val), NULL);
+}
 
-	tm_now(&now);
-	gnet_stats_crc32 = crc32_update(gnet_stats_crc32, &now, sizeof now);
-	gnet_host_set(&host, n->addr, n->port);
-	gnet_stats_crc32 = crc32_update(
-		gnet_stats_crc32, &host, gnet_host_length(&host));
-	gnet_stats_crc32 = crc32_update(gnet_stats_crc32, &type, sizeof type);
-	gnet_stats_crc32 = crc32_update(gnet_stats_crc32, &val, sizeof val);
+static void
+gnet_stats_count_received_header_internal(gnutella_node_t *n,
+	size_t header_size, uint t, uint8 ttl, uint8 hops)
+{
+	gnet_stats_t *stats = NODE_USES_UDP(n) ? &gnet_udp_stats : &gnet_tcp_stats;
+	uint i;
+
+    n->received++;
+
+    gnet_stats.pkg.received[MSG_TOTAL]++;
+    gnet_stats.pkg.received[t]++;
+    gnet_stats.byte.received[MSG_TOTAL] += header_size;
+    gnet_stats.byte.received[t] += header_size;
+
+    stats->pkg.received[MSG_TOTAL]++;
+    stats->pkg.received[t]++;
+    stats->byte.received[MSG_TOTAL] += header_size;
+    stats->byte.received[t] += header_size;
+
+	i = MIN(ttl, STATS_RECV_COLUMNS - 1);
+    stats->pkg.received_ttl[i][MSG_TOTAL]++;
+    stats->pkg.received_ttl[i][t]++;
+
+	i = MIN(hops, STATS_RECV_COLUMNS - 1);
+    stats->pkg.received_hops[i][MSG_TOTAL]++;
+    stats->pkg.received_hops[i][t]++;
 }
 
 /**
@@ -514,30 +264,15 @@ void
 gnet_stats_count_received_header(gnutella_node_t *n)
 {
 	uint t = stats_lut[gnutella_header_get_function(&n->header)];
-	uint i;
-	gnet_stats_t *stats;
+	uint8 ttl, hops;
 
-	stats = NODE_USES_UDP(n) ? &gnet_udp_stats : &gnet_tcp_stats;
+	g_assert(thread_is_main());
+	g_assert(!NODE_TALKS_G2(n));
 
-    n->received++;
+	ttl = gnutella_header_get_ttl(&n->header);
+	hops = gnutella_header_get_hops(&n->header);
 
-    gnet_stats.pkg.received[MSG_TOTAL]++;
-    gnet_stats.pkg.received[t]++;
-    gnet_stats.byte.received[MSG_TOTAL] += GTA_HEADER_SIZE;
-    gnet_stats.byte.received[t] += GTA_HEADER_SIZE;
-
-    stats->pkg.received[MSG_TOTAL]++;
-    stats->pkg.received[t]++;
-    stats->byte.received[MSG_TOTAL] += GTA_HEADER_SIZE;
-    stats->byte.received[t] += GTA_HEADER_SIZE;
-
-	i = MIN(gnutella_header_get_ttl(&n->header), STATS_RECV_COLUMNS - 1);
-    stats->pkg.received_ttl[i][MSG_TOTAL]++;
-    stats->pkg.received_ttl[i][t]++;
-
-	i = MIN(gnutella_header_get_hops(&n->header), STATS_RECV_COLUMNS - 1);
-    stats->pkg.received_hops[i][MSG_TOTAL]++;
-    stats->pkg.received_hops[i][t]++;
+	gnet_stats_count_received_header_internal(n, GTA_HEADER_SIZE, t, ttl, hops);
 }
 
 /**
@@ -553,6 +288,9 @@ gnet_stats_count_kademlia_header(const gnutella_node_t *n, uint kt)
 	uint i;
 	gnet_stats_t *stats;
 
+	g_assert(thread_is_main());
+	g_assert(!NODE_TALKS_G2(n));
+
 	stats = NODE_USES_UDP(n) ? &gnet_udp_stats : &gnet_tcp_stats;
 
     gnet_stats.pkg.received[t]--;
@@ -582,21 +320,28 @@ gnet_stats_count_kademlia_header(const gnutella_node_t *n, uint kt)
 }
 
 /**
- * Called when Gnutella payload has been read.
+ * Called when Gnutella payload has been read, or when a G2 messsage is read.
  *
  * The actual payload size (effectively read) is expected to be found
- * in n->size.
+ * in n->size for Gnutella messages and G2 messages.
+ *
+ * @param n			the node from which message was received
+ * @param payload	start of Gnutella payload, or head of G2 frame
  */
 void
 gnet_stats_count_received_payload(const gnutella_node_t *n, const void *payload)
 {
-	uint8 f = gnutella_header_get_function(&n->header);
-	uint t = stats_lut[f];
+	uint8 f;
+	uint t;
 	uint i;
 	gnet_stats_t *stats;
     uint32 size;
+	uint8 hops, ttl;
+
+	g_assert(thread_is_main());
 
 	stats = NODE_USES_UDP(n) ? &gnet_udp_stats : &gnet_tcp_stats;
+	size = n->size;
 
 	/*
 	 * Size is NOT read in the Gnutella header but in n->size, which
@@ -609,7 +354,30 @@ gnet_stats_count_received_payload(const gnutella_node_t *n, const void *payload)
 	 *		--RAM, 2010-10-30
 	 */
 
-	size = n->size;
+	if (NODE_TALKS_G2(n)) {
+		f = g2_msg_type(payload, size);
+		if (f != G2_MSG_MAX) {
+			f += MSG_G2_BASE;
+		} else {
+			f = G_N_ELEMENTS(stats_lut) - 1;	/* Last, holds MSG_UNKNOWN */
+		}
+		ttl = 1;
+		hops = NODE_USES_UDP(n) ? 0 : 1;
+		t = stats_lut[f];
+		/*
+		 * No header for G2, so count header reception now with a size of zero.
+		 * This is required to update the other packet reception statistics.
+		 */
+		gnet_stats_count_received_header_internal(
+			deconstify_pointer(n),
+			0, t, ttl, hops);
+	} else {
+		f = gnutella_header_get_function(&n->header);
+		hops = gnutella_header_get_hops(&n->header);
+		ttl = gnutella_header_get_ttl(&n->header);
+		t = stats_lut[f];
+	}
+
 	gnet_stats_randomness(n, f, size);
 
 	/*
@@ -633,49 +401,31 @@ gnet_stats_count_received_payload(const gnutella_node_t *n, const void *payload)
 		}
 	}
 
+	g_assert(t < MSG_TOTAL);
+
     gnet_stats.byte.received[MSG_TOTAL] += size;
     gnet_stats.byte.received[t] += size;
 
     stats->byte.received[MSG_TOTAL] += size;
     stats->byte.received[t] += size;
 
-	i = MIN(gnutella_header_get_ttl(&n->header), STATS_RECV_COLUMNS - 1);
+	i = MIN(ttl, STATS_RECV_COLUMNS - 1);
     stats->byte.received_ttl[i][MSG_TOTAL] += size;
     stats->byte.received_ttl[i][t] += size;
 
-	i = MIN(gnutella_header_get_hops(&n->header), STATS_RECV_COLUMNS - 1);
+	i = MIN(hops, STATS_RECV_COLUMNS - 1);
     stats->byte.received_hops[i][MSG_TOTAL] += size;
     stats->byte.received_hops[i][t] += size;
 }
 
-void
-gnet_stats_count_queued(const gnutella_node_t *n,
-	uint8 type, const void *base, uint32 size)
+static void
+gnet_stats_count_queued_internal(const gnutella_node_t *n,
+	uint t, uint8 hops, uint32 size, gnet_stats_t *stats)
 {
 	uint64 *stats_pkg;
 	uint64 *stats_byte;
-	uint t = stats_lut[type];
-	gnet_stats_t *stats;
-	uint8 hops;
 
-	g_assert(t != MSG_UNKNOWN);
-
-	stats = NODE_USES_UDP(n) ? &gnet_udp_stats : &gnet_tcp_stats;
-
-	/*
-	 * Adjust for Kademlia messages.
-	 */
-
-	if (GTA_MSG_DHT == type && size >= KDA_HEADER_SIZE) {
-		uint8 opcode = kademlia_header_get_function(base);
-
-		if (UNSIGNED(opcode + MSG_DHT_BASE) < G_N_ELEMENTS(stats_lut)) {
-			t = stats_lut[opcode + MSG_DHT_BASE];
-		}
-		hops = 0;
-	} else {
-		hops = gnutella_header_get_hops(base);
-	}
+	g_assert(t < MSG_TOTAL);
 
 	gnet_stats_randomness(n, t & 0xff, size);
 
@@ -697,16 +447,16 @@ gnet_stats_count_queued(const gnutella_node_t *n,
 }
 
 void
-gnet_stats_count_sent(const gnutella_node_t *n,
+gnet_stats_count_queued(const gnutella_node_t *n,
 	uint8 type, const void *base, uint32 size)
 {
-	uint64 *stats_pkg;
-	uint64 *stats_byte;
 	uint t = stats_lut[type];
 	gnet_stats_t *stats;
 	uint8 hops;
 
 	g_assert(t != MSG_UNKNOWN);
+	g_assert(thread_is_main());
+	g_assert(!NODE_TALKS_G2(n));
 
 	stats = NODE_USES_UDP(n) ? &gnet_udp_stats : &gnet_tcp_stats;
 
@@ -725,6 +475,45 @@ gnet_stats_count_sent(const gnutella_node_t *n,
 		hops = gnutella_header_get_hops(base);
 	}
 
+	gnet_stats_count_queued_internal(n, t, hops, size, stats);
+}
+
+void
+gnet_stats_g2_count_queued(const gnutella_node_t *n,
+	const void *base, size_t len)
+{
+	gnet_stats_t *stats;
+	uint t;
+	uint8 f;
+
+	g_assert(thread_is_main());
+	g_assert(NODE_TALKS_G2(n));
+
+	stats = NODE_USES_UDP(n) ? &gnet_udp_stats : &gnet_tcp_stats;
+
+	f = g2_msg_type(base, len);
+
+	if (f != G2_MSG_MAX) {
+		f += MSG_G2_BASE;
+	} else {
+		f = G_N_ELEMENTS(stats_lut) - 1;	/* Last, holds MSG_UNKNOWN */
+	}
+
+	t = stats_lut[f];
+
+	/* Leaf mode => hops = 0 */
+	gnet_stats_count_queued_internal(n, t, 0, len, stats);
+}
+
+static void
+gnet_stats_count_sent_internal(const gnutella_node_t *n,
+	uint t, uint8 hops, uint32 size, gnet_stats_t *stats)
+{
+	uint64 *stats_pkg;
+	uint64 *stats_byte;
+
+	g_assert(t < MSG_TOTAL);
+
 	gnet_stats_randomness(n, t & 0xff, size);
 
 	stats_pkg = hops ? gnet_stats.pkg.relayed : gnet_stats.pkg.generated;
@@ -744,6 +533,59 @@ gnet_stats_count_sent(const gnutella_node_t *n,
     stats_byte[t] += size;
 }
 
+void
+gnet_stats_count_sent(const gnutella_node_t *n,
+	uint8 type, const void *base, uint32 size)
+{
+	uint t = stats_lut[type];
+	gnet_stats_t *stats;
+	uint8 hops;
+
+	g_assert(t != MSG_UNKNOWN);
+	g_assert(thread_is_main());
+	g_assert(!NODE_TALKS_G2(n));
+
+	stats = NODE_USES_UDP(n) ? &gnet_udp_stats : &gnet_tcp_stats;
+
+	/*
+	 * Adjust for Kademlia messages.
+	 */
+
+	if (GTA_MSG_DHT == type && size >= KDA_HEADER_SIZE) {
+		uint8 opcode = kademlia_header_get_function(base);
+
+		if (UNSIGNED(opcode + MSG_DHT_BASE) < G_N_ELEMENTS(stats_lut)) {
+			t = stats_lut[opcode + MSG_DHT_BASE];
+		}
+		hops = 0;
+	} else {
+		hops = gnutella_header_get_hops(base);
+	}
+
+	gnet_stats_count_sent_internal(n, t, hops, size, stats);
+}
+
+void
+gnet_stats_g2_count_sent(const gnutella_node_t *n,
+	enum g2_msg type, uint32 size)
+{
+	uint t;
+	gnet_stats_t *stats;
+
+	g_assert(thread_is_main());
+	g_assert((uint) type < UNSIGNED(G2_MSG_MAX));
+	g_assert(NODE_TALKS_G2(n));
+
+	stats = NODE_USES_UDP(n) ? &gnet_udp_stats : &gnet_tcp_stats;
+
+	t = stats_lut[MSG_G2_BASE + type];
+
+	g_assert(t != MSG_UNKNOWN);
+
+	/* Leaf mode => hops = 0 */
+	gnet_stats_count_sent_internal(n, t, 0, size, stats);
+}
+
 void
 gnet_stats_count_expired(const gnutella_node_t *n)
 {
@@ -751,6 +593,9 @@ gnet_stats_count_expired(const gnutella_node_t *n)
 	uint t = stats_lut[gnutella_header_get_function(&n->header)];
 	gnet_stats_t *stats;
 
+	g_assert(thread_is_main());
+	g_assert(!NODE_TALKS_G2(n));
+
 	stats = NODE_USES_UDP(n) ? &gnet_udp_stats : &gnet_tcp_stats;
 
     gnet_stats.pkg.expired[MSG_TOTAL]++;
@@ -777,6 +622,8 @@ gnet_stats_count_expired(const gnutella_node_t *n)
     gnet_stats.pkg.dropped[t]++;						\
     gnet_stats.byte.dropped[MSG_TOTAL] += (s);			\
     gnet_stats.byte.dropped[t] += (s);					\
+	gs->drop_reason[reason][MSG_TOTAL]++;				\
+	gs->drop_reason[reason][t]++;						\
     gs->pkg.dropped[MSG_TOTAL]++;						\
     gs->pkg.dropped[t]++;								\
     gs->byte.dropped[MSG_TOTAL] += (s);					\
@@ -791,12 +638,27 @@ gnet_stats_count_dropped(gnutella_node_t *n, msg_drop_reason_t reason)
 	gnet_stats_t *stats;
 
 	g_assert(UNSIGNED(reason) < MSG_DROP_REASON_COUNT);
+	g_assert(thread_is_main());
 
-    size = n->size + sizeof(n->header);
-	type = stats_lut[gnutella_header_get_function(&n->header)];
 	stats = NODE_USES_UDP(n) ? &gnet_udp_stats : &gnet_tcp_stats;
 
-	gnet_stats_randomness(n, type & 0xff, size);
+	if (NODE_TALKS_G2(n)) {
+		int f = g2_msg_type(n->data, n->size);
+		if (f != G2_MSG_MAX) {
+			f += MSG_G2_BASE;
+		} else {
+			f = G_N_ELEMENTS(stats_lut) - 1;	/* Last, holds MSG_UNKNOWN */
+		}
+		type = stats_lut[f];
+		size = n->size;
+	} else {
+		type = stats_lut[gnutella_header_get_function(&n->header)];
+		size = n->size + sizeof(n->header);
+	}
+
+	entropy_harvest_small(
+		VARLEN(n->addr), VARLEN(n->port), VARLEN(reason), VARLEN(type),
+		VARLEN(size), NULL);
 
 	DROP_STATS(stats, type, size);
 	node_inc_rxdrop(n);
@@ -808,10 +670,19 @@ gnet_stats_count_dropped(gnutella_node_t *n, msg_drop_reason_t reason)
 	default: ;
 	}
 
-	if (GNET_PROPERTY(log_dropped_gnutella))
-		gmsg_log_split_dropped(&n->header, n->data, n->size,
-			"from %s: %s", node_infostr(n),
-			gnet_stats_drop_reason_to_string(reason));
+	if (NODE_TALKS_G2(n)) {
+		if (GNET_PROPERTY(log_dropped_g2)) {
+			g2_msg_log_dropped_data(n->data, n->size,
+				"from %s: %s", node_infostr(n),
+				gnet_stats_drop_reason_to_string(reason));
+		}
+	} else {
+		if (GNET_PROPERTY(log_dropped_gnutella)) {
+			gmsg_log_split_dropped(&n->header, n->data, n->size,
+				"from %s: %s", node_infostr(n),
+				gnet_stats_drop_reason_to_string(reason));
+		}
+	}
 }
 
 /**
@@ -828,12 +699,15 @@ gnet_dht_stats_count_dropped(gnutella_node_t *n, kda_msg_t opcode,
 	g_assert(UNSIGNED(reason) < MSG_DROP_REASON_COUNT);
 	g_assert(opcode <= KDA_MSG_MAX_ID);
 	g_assert(UNSIGNED(opcode + MSG_DHT_BASE) < G_N_ELEMENTS(stats_lut));
+	g_assert(thread_is_main());
 
     size = n->size + sizeof(n->header);
 	type = stats_lut[opcode + MSG_DHT_BASE];
 	stats = NODE_USES_UDP(n) ? &gnet_udp_stats : &gnet_tcp_stats;
 
-	gnet_stats_randomness(n, type & 0xff, size);
+	entropy_harvest_small(
+		VARLEN(n->addr), VARLEN(n->port), VARLEN(reason), VARLEN(type),
+		VARLEN(size), NULL);
 
 	DROP_STATS(stats, type, size);
 	node_inc_rxdrop(n);
@@ -848,7 +722,10 @@ gnet_stats_count_general(gnr_stats_t type, int delta)
 	size_t i = type;
 
 	g_assert(i < GNR_TYPE_COUNT);
+
+	GNET_STATS_LOCK;
     gnet_stats.general[i] += delta;
+	GNET_STATS_UNLOCK;
 }
 
 /**
@@ -860,7 +737,10 @@ gnet_stats_inc_general(gnr_stats_t type)
 	size_t i = type;
 
 	g_assert(i < GNR_TYPE_COUNT);
+
+	GNET_STATS_LOCK;
     gnet_stats.general[i]++;
+	GNET_STATS_UNLOCK;
 }
 
 /**
@@ -872,7 +752,10 @@ gnet_stats_dec_general(gnr_stats_t type)
 	size_t i = type;
 
 	g_assert(i < GNR_TYPE_COUNT);
+
+	GNET_STATS_LOCK;
     gnet_stats.general[i]--;
+	GNET_STATS_UNLOCK;
 }
 
 /**
@@ -884,8 +767,11 @@ gnet_stats_max_general(gnr_stats_t type, uint64 value)
 	size_t i = type;
 
 	g_assert(i < GNR_TYPE_COUNT);
+
+	GNET_STATS_LOCK;
 	if (value > gnet_stats.general[i])
 		gnet_stats.general[i] = value;
+	GNET_STATS_UNLOCK;
 }
 
 /**
@@ -897,7 +783,10 @@ gnet_stats_set_general(gnr_stats_t type, uint64 value)
 	size_t i = type;
 
 	g_assert(i < GNR_TYPE_COUNT);
+
+	GNET_STATS_LOCK;
     gnet_stats.general[i] = value;
+	GNET_STATS_UNLOCK;
 }
 
 /**
@@ -907,9 +796,15 @@ uint64
 gnet_stats_get_general(gnr_stats_t type)
 {
 	size_t i = type;
+	uint64 value;
 
 	g_assert(i < GNR_TYPE_COUNT);
-	return gnet_stats.general[i];
+
+	GNET_STATS_LOCK;
+	value = gnet_stats.general[i];
+	GNET_STATS_UNLOCK;
+
+	return value;
 }
 
 void
@@ -920,10 +815,14 @@ gnet_stats_count_dropped_nosize(
 	gnet_stats_t *stats;
 
 	g_assert(UNSIGNED(reason) < MSG_DROP_REASON_COUNT);
+	g_assert(thread_is_main());
+	g_assert(!NODE_TALKS_G2(n));
 
 	type = stats_lut[gnutella_header_get_function(&n->header)];
 	stats = NODE_USES_UDP(n) ? &gnet_udp_stats : &gnet_tcp_stats;
 
+	entropy_harvest_small(VARLEN(n->addr), VARLEN(n->port), NULL);
+
 	/* Data part of message not read */
 	DROP_STATS(stats, type, sizeof(n->header));
 
@@ -933,16 +832,44 @@ gnet_stats_count_dropped_nosize(
 			gnet_stats_drop_reason_to_string(reason));
 }
 
+static void
+gnet_stats_flowc_internal(uint t,
+	uint8 function, uint8 ttl, uint8 hops, size_t size)
+{
+	uint i;
+
+	g_assert(t < MSG_TOTAL);
+
+	i = MIN(hops, STATS_FLOWC_COLUMNS - 1);
+	gnet_stats.pkg.flowc_hops[i][t]++;
+	gnet_stats.pkg.flowc_hops[i][MSG_TOTAL]++;
+	gnet_stats.byte.flowc_hops[i][t] += size;
+	gnet_stats.byte.flowc_hops[i][MSG_TOTAL] += size;
+
+	i = MIN(ttl, STATS_FLOWC_COLUMNS - 1);
+
+	/* Cannot send a message with TTL=0 (DHT messages are not Gnutella) */
+	g_assert(function == GTA_MSG_DHT || i != 0);
+
+	gnet_stats.pkg.flowc_ttl[i][t]++;
+	gnet_stats.pkg.flowc_ttl[i][MSG_TOTAL]++;
+	gnet_stats.byte.flowc_ttl[i][t] += size;
+	gnet_stats.byte.flowc_ttl[i][MSG_TOTAL] += size;
+
+	entropy_harvest_small(VARLEN(t), VARLEN(function), VARLEN(size), NULL);
+}
+
 void
 gnet_stats_count_flowc(const void *head, bool head_only)
 {
 	uint t;
-	uint i;
 	uint16 size = gmsg_size(head) + GTA_HEADER_SIZE;
 	uint8 function = gnutella_header_get_function(head);
 	uint8 ttl = gnutella_header_get_ttl(head);
 	uint8 hops = gnutella_header_get_hops(head);
 
+	g_assert(thread_is_main());
+
 	if (GNET_PROPERTY(node_debug) > 3)
 		g_debug("FLOWC function=%d ttl=%d hops=%d", function, ttl, hops);
 
@@ -964,21 +891,35 @@ gnet_stats_count_flowc(const void *head, bool head_only)
 		t = stats_lut[function];
 	}
 
-	i = MIN(hops, STATS_FLOWC_COLUMNS - 1);
-	gnet_stats.pkg.flowc_hops[i][t]++;
-	gnet_stats.pkg.flowc_hops[i][MSG_TOTAL]++;
-	gnet_stats.byte.flowc_hops[i][t] += size;
-	gnet_stats.byte.flowc_hops[i][MSG_TOTAL] += size;
+	gnet_stats_flowc_internal(t, function, ttl, hops, size);
+}
 
-	i = MIN(ttl, STATS_FLOWC_COLUMNS - 1);
+void
+gnet_stats_g2_count_flowc(const gnutella_node_t *n,
+	const void *base, size_t len)
+{
+	uint t;
+	uint8 f, ttl, hops;
 
-	/* Cannot send a message with TTL=0 (DHT messages are not Gnutella) */
-	g_assert(function == GTA_MSG_DHT || i != 0);
+	g_assert(thread_is_main());
 
-	gnet_stats.pkg.flowc_ttl[i][t]++;
-	gnet_stats.pkg.flowc_ttl[i][MSG_TOTAL]++;
-	gnet_stats.byte.flowc_ttl[i][t] += size;
-	gnet_stats.byte.flowc_ttl[i][MSG_TOTAL] += size;
+	f = g2_msg_type(base, len);
+
+	if (GNET_PROPERTY(node_debug) > 3)
+		g_debug("FLOWC G2 %s", g2_msg_type_name(f));
+
+	if (f != G2_MSG_MAX) {
+		f += MSG_G2_BASE;
+	} else {
+		f = G_N_ELEMENTS(stats_lut) - 1;	/* Last, holds MSG_UNKNOWN */
+	}
+
+	ttl = NODE_USES_UDP(n) ? 1 : 2;		/* Purely made up, but cannot be 0 */
+	hops = 0;		/* Locally generated, this is TX flowc */
+
+	t = stats_lut[f];
+
+	gnet_stats_flowc_internal(t, f, ttl, hops, len);
 }
 
 /***
@@ -989,13 +930,18 @@ void
 gnet_stats_get(gnet_stats_t *s)
 {
     g_assert(s != NULL);
+
+	GNET_STATS_LOCK;
     *s = gnet_stats;
+	GNET_STATS_UNLOCK;
 }
 
 void
 gnet_stats_tcp_get(gnet_stats_t *s)
 {
     g_assert(s != NULL);
+	g_assert(thread_is_main());
+
     *s = gnet_tcp_stats;
 }
 
@@ -1003,6 +949,8 @@ void
 gnet_stats_udp_get(gnet_stats_t *s)
 {
     g_assert(s != NULL);
+	g_assert(thread_is_main());
+
     *s = gnet_udp_stats;
 }
 
