@@ -34,6 +34,7 @@
 #include "common.h"
 
 #include "pproxy.h"
+
 #include "http.h"
 #include "hosts.h"
 #include "version.h"
@@ -58,6 +59,7 @@
 
 #include "downloads.h"
 #include "features.h"
+#include "hosts.h"
 #include "settings.h"			/* For listen_addr() */
 #include "token.h"
 
@@ -66,14 +68,16 @@
 #include "lib/concat.h"
 #include "lib/endian.h"
 #include "lib/getline.h"
-#include "lib/glib-missing.h"
 #include "lib/halloc.h"
 #include "lib/hashlist.h"
 #include "lib/header.h"
+#include "lib/host_addr.h"
 #include "lib/log.h"
 #include "lib/parse.h"
+#include "lib/pslist.h"
 #include "lib/sequence.h"
 #include "lib/str.h"
+#include "lib/stringify.h"
 #include "lib/tm.h"
 #include "lib/unsigned.h"
 #include "lib/walloc.h"
@@ -86,7 +90,7 @@
  *** Server-side of push-proxy
  ***/
 
-static GSList *pproxies = NULL;	/**< Currently active push-proxy requests */
+static pslist_t *pproxies = NULL;	/**< Currently active push-proxy requests */
 
 static void send_pproxy_error(struct pproxy *pp, int code,
 	const char *msg, ...) G_GNUC_PRINTF(3, 4);
@@ -220,7 +224,7 @@ pproxy_remove_v(struct pproxy *pp, const char *reason, va_list ap)
 	pp->magic = 0;
 	WFREE(pp);
 
-	pproxies = g_slist_remove(pproxies, pp);
+	pproxies = pslist_remove(pproxies, pp);
 }
 
 /**
@@ -267,10 +271,10 @@ pproxy_error_remove(struct pproxy *pp, int code, const char *msg, ...)
 void
 pproxy_timer(time_t now)
 {
-	GSList *sl;
-	GSList *to_remove = NULL;
+	pslist_t *sl;
+	pslist_t *to_remove = NULL;
 
-	for (sl = pproxies; sl; sl = g_slist_next(sl)) {
+	PSLIST_FOREACH(pproxies, sl) {
 		struct pproxy *pp = sl->data;
 
 		pproxy_check(pp);
@@ -284,16 +288,16 @@ pproxy_timer(time_t now)
 			delta_time(now, pp->last_update) >
 				(time_delta_t) GNET_PROPERTY(upload_connecting_timeout)
 		) {
-			to_remove = g_slist_prepend(to_remove, pp);
+			to_remove = pslist_prepend(to_remove, pp);
 		}
 	}
 
-	for (sl = to_remove; sl; sl = g_slist_next(sl)) {
+	PSLIST_FOREACH(to_remove, sl) {
 		struct pproxy *pp = sl->data;
 		pproxy_error_remove(pp, 408, "Request timeout");
 	}
 
-	g_slist_free(to_remove);
+	pslist_free(to_remove);
 }
 
 /**
@@ -409,7 +413,7 @@ get_params(struct pproxy *pp, const char *request,
 	if (!up) {
 		pproxy_error_remove(pp, 400,
 			"Malformed push-proxy request: Bad URL encoding");
-		goto error;
+		return FALSE;
 	}
 	value = url_params_get(up, attr);
 
@@ -426,6 +430,7 @@ get_params(struct pproxy *pp, const char *request,
 	datalen = strlen(value);
 
 	if (0 == strcmp(attr, "ServerId")) {
+		struct guid buf;
 		const struct guid *guid;
 
 		/*
@@ -434,15 +439,15 @@ get_params(struct pproxy *pp, const char *request,
 
 		if (datalen != 26 && datalen != 32) {
 			pproxy_error_remove(pp, 400, "Malformed push-proxy request: "
-				"wrong length for parameter \"%s\": %d byte%s", attr, datalen,
-				datalen == 1 ? "" : "s");
+				"wrong length for parameter \"%s\": %d byte%s",
+				attr, datalen, plural(datalen));
 			goto error;
 		}
 
 		if (GNET_PROPERTY(push_proxy_debug) > 0)
 			g_debug("PUSH-PROXY: decoding %s=%s as base32", attr, value);
 
-		guid = base32_to_guid(value);
+		guid = base32_to_guid(value, &buf);
 		if (guid == NULL) {
 			pproxy_error_remove(pp, 400, "Malformed push-proxy request: "
 				"parameter \"%s\" is not valid base32", attr);
@@ -459,8 +464,8 @@ get_params(struct pproxy *pp, const char *request,
 
 		if (datalen != 32) {
 			pproxy_error_remove(pp, 400, "Malformed push-proxy request: "
-				"wrong length for parameter \"%s\": %d byte%s", attr, datalen,
-				datalen == 1 ? "" : "s");
+				"wrong length for parameter \"%s\": %d byte%s",
+				attr, datalen, plural(datalen));
 			goto error;
 		}
 
@@ -562,6 +567,8 @@ build_push(uint8 ttl, uint8 hops, const struct guid *guid,
 	 */
 
 	primary = is_host_addr(addr_v4) ? addr_v4 : addr_v6;
+	if (!is_host_addr(primary))
+		return zero_array;
 
 	ggep_stream_init(&gs, p, size);
 
@@ -716,11 +723,11 @@ pproxy_request(struct pproxy *pp, header_t *header)
 {
 	struct gnutella_socket *s = pp->socket;
 	const char *request = getline_str(s->getline);
-	struct gnutella_node *n;
+	gnutella_node_t *n;
 	const char *buf;
 	char *token;
 	char *user_agent;
-	GSList *nodes;
+	pslist_t *nodes;
 	bool supports_tls = FALSE;
 
 	if (GNET_PROPERTY(push_proxy_trace) & SOCK_TRACE_IN) {
@@ -789,8 +796,8 @@ pproxy_request(struct pproxy *pp, header_t *header)
 	}
 
 	if (
-		!host_addr_equal(pp->addr_v4, s->addr) &&
-		!host_addr_equal(pp->addr_v6, s->addr)
+		!host_addr_equiv(pp->addr_v4, s->addr) &&
+		!host_addr_equiv(pp->addr_v6, s->addr)
 	) {
 		g_warning("push-proxy request from %s (%s) said node was at %s/%s",
 			host_addr_to_string(s->addr), pproxy_vendor_str(pp),
@@ -874,23 +881,23 @@ pproxy_request(struct pproxy *pp, header_t *header)
 				host_addr_port_to_string2(pp->addr_v6, pp->port),
 				(ulong) pp->file_idx);
 		} else {
-			int cnt;
+			size_t cnt;
 
 			gmsg_sendto_all(nodes, packet.data, packet.size);
 			gnet_stats_inc_general(GNR_PUSH_PROXY_BROADCASTED);
 
-			cnt = g_slist_length(nodes);
+			cnt = pslist_length(nodes);
 
 			http_send_status(HTTP_PUSH_PROXY, pp->socket, 203, FALSE, NULL, 0,
-					"Push-proxy: message sent through Gnutella (via %d node%s)",
-					cnt, cnt == 1 ? "" : "s");
+					"Push-proxy: message sent through Gnutella "
+					"(via %zd node%s)", cnt, plural(cnt));
 
 			pp->error_sent = 203;
-			pproxy_remove(pp, "Push sent via Gnutella (%d node%s) for GUID %s",
-					cnt, cnt == 1 ? "" : "s", guid_hex_str(pp->guid));
+			pproxy_remove(pp, "Push sent via Gnutella (%zd node%s) for GUID %s",
+					cnt, plural(cnt), guid_hex_str(pp->guid));
 		}
 
-		gm_slist_free_null(&nodes);
+		pslist_free_null(&nodes);
 		return;
 	}
 
@@ -1053,9 +1060,9 @@ pproxy_add(struct gnutella_socket *s)
 void
 pproxy_close(void)
 {
-	GSList *l;
+	pslist_t *l;
 
-	for (l = pproxies; l; l = g_slist_next(l)) {
+	PSLIST_FOREACH(pproxies, l) {
 		struct pproxy *pp = l->data;
 
 		pproxy_free_resources(pp);
@@ -1063,7 +1070,7 @@ pproxy_close(void)
 		WFREE(pp);
 	}
 
-	gm_slist_free_null(&pproxies);
+	pslist_free_null(&pproxies);
 }
 
 /***
@@ -1338,14 +1345,13 @@ cproxy_http_newstate(struct http_async *handle, http_state_t newstate)
 }
 
 static void
-cproxy_http_start(cqueue_t *unused_cq, void *obj)
+cproxy_http_start(cqueue_t *cq, void *obj)
 {
 	struct cproxy *cp = obj;
 
-	(void) unused_cq;
 	cproxy_check(cp);
 
-	cp->udp_ev = NULL;
+	cq_zero(cq, &cp->udp_ev);
 	cproxy_http_request(cp);
 }
 
@@ -1562,7 +1568,7 @@ pproxy_set_allocate(size_t max_proxies)
 
 	WALLOC0(ps);
 	ps->magic = PPROXY_SET_MAGIC;
-	ps->proxies = hash_list_new(gnet_host_hash, gnet_host_eq);
+	ps->proxies = hash_list_new(gnet_host_hash, gnet_host_equal);
 	ps->max_proxies = max_proxies;
 
 	return ps;
@@ -1606,7 +1612,8 @@ pproxy_set_trim(const pproxy_set_t *ps)
 /**
  * Add a push-proxy to the set.
  *
- * @return TRUE if host was added, FALSE if we already knew it.
+ * @return TRUE if host was added, FALSE if we already knew it or the
+ * host was invalid.
  */
 bool
 pproxy_set_add(pproxy_set_t *ps, const host_addr_t addr, uint16 port)
@@ -1616,7 +1623,11 @@ pproxy_set_add(pproxy_set_t *ps, const host_addr_t addr, uint16 port)
 
 	pproxy_set_check(ps);
 
+	if (!host_is_valid(addr, port))
+		return FALSE;
+
 	gnet_host_set(&host, addr, port);
+
 	if (hash_list_contains(ps->proxies, &host)) {
 		hash_list_moveto_head(ps->proxies, &host);
 	} else {
@@ -1642,6 +1653,12 @@ pproxy_set_add_vec(pproxy_set_t *ps, const gnet_host_vec_t *vec)
 
 	for (i = gnet_host_vec_count(vec) - 1; i >= 0; i--) {
 		gnet_host_t host = gnet_host_vec_get(vec, i);
+		host_addr_t addr = gnet_host_get_addr(&host);
+		uint16 port = gnet_host_get_port(&host);
+
+		if (!host_is_valid(addr, port))
+			continue;
+
 		if (hash_list_contains(ps->proxies, &host)) {
 			hash_list_moveto_head(ps->proxies, &host);
 		} else {
@@ -1664,10 +1681,17 @@ pproxy_set_add_array(pproxy_set_t *ps, gnet_host_t *proxies, int proxy_count)
 	pproxy_set_check(ps);
 
 	for (i = 0; i < proxy_count; i++) {
-		if (hash_list_contains(ps->proxies, &proxies[i])) {
-			hash_list_moveto_head(ps->proxies, &proxies[i]);
+		gnet_host_t *host = &proxies[i];
+		host_addr_t addr = gnet_host_get_addr(host);
+		uint16 port = gnet_host_get_port(host);
+
+		if (!host_is_valid(addr, port))
+			continue;
+
+		if (hash_list_contains(ps->proxies, host)) {
+			hash_list_moveto_head(ps->proxies, host);
 		} else {
-			hash_list_prepend(ps->proxies, gnet_host_dup(&proxies[i]));
+			hash_list_prepend(ps->proxies, gnet_host_dup(host));
 		}
 	}
 
