@@ -55,6 +55,7 @@
 #include "lib/parse.h"
 #include "lib/random.h"
 #include "lib/str.h"
+#include "lib/stringify.h"
 #include "lib/walloc.h"
 
 #include "if/gnet_property_priv.h"
@@ -101,8 +102,9 @@ static const struct {
 #else	/* !USE_LOCAL_UHC */
 	{ "1.uhc.gtk-gnutella.nl:19104" },
 	{ "uhc.gtk-gnutella.nl:15749" },
-	{ "g1.uswest.dyslexicfish.net:33558" },
-	{ "g1.uk.dyslexicfish.net:33558" },
+	{ "useast.gnutella.dyslexicfish.net:3558" },
+	{ "uswest.gnutella.dyslexicfish.net:3558" },
+	{ "uk.gnutella.dyslexicfish.net:3558" },
 #endif	/* USE_LOCAL_UHC */
 };
 
@@ -344,16 +346,15 @@ uhc_try_random(void)
  * get a reply within the specified timeout.
  */
 static void
-uhc_ping_timeout(cqueue_t *unused_cq, void *unused_obj)
+uhc_ping_timeout(cqueue_t *cq, void *unused_obj)
 {
-	(void) unused_cq;
 	(void) unused_obj;
 
 	if (GNET_PROPERTY(bootstrap_debug))
 		g_warning("no reply from UDP host cache %s:%u",
 			uhc_ctx.host, uhc_ctx.port);
 
-	uhc_ctx.timeout_ev = NULL;
+	cq_zero(cq, &uhc_ctx.timeout_ev);
 	uhc_try_random();
 }
 
@@ -530,7 +531,7 @@ uhc_ipp_extract(gnutella_node_t *n, const char *payload, int paylen,
 
 	if (GNET_PROPERTY(bootstrap_debug))
 		g_debug("extracting %d host%s in UDP IPP pong #%s from %s",
-			cnt, cnt == 1 ? "" : "s",
+			cnt, plural(cnt),
 			guid_hex_str(gnutella_header_get_muid(&n->header)), node_addr(n));
 
 	for (i = 0, p = payload; i < cnt; i++, p = const_ptr_add_offset(p, len)) {
@@ -561,7 +562,7 @@ uhc_ipp_extract(gnutella_node_t *n, const char *payload, int paylen,
 
 	if (GNET_PROPERTY(bootstrap_debug)) {
 		g_debug("BOOT UDP cache \"%s\" replied: got %d host%s from %s",
-			uhc_ctx.host, cnt, cnt == 1 ? "" : "s", node_addr(n));
+			uhc_ctx.host, cnt, plural(cnt), node_addr(n));
 	}
 
 	/*
