@@ -44,6 +44,7 @@
 #include "lib/hashing.h"
 #include "lib/host_addr.h"
 #include "lib/htable.h"
+#include "lib/stringify.h"
 #include "lib/walloc.h"
 
 #include "lib/override.h"		/* Must be the last header included */
@@ -142,14 +143,13 @@ urpc_received(const gnutella_socket_t *s,
  * RPC timed out.
  */
 static void
-urpc_timed_out(cqueue_t *unused_cq, void *obj)
+urpc_timed_out(cqueue_t *cq, void *obj)
 {
 	struct urpc_cb *ucb = obj;
 
 	urpc_cb_check(ucb);
-	(void) unused_cq;
 
-	ucb->timeout_ev = NULL;
+	cq_zero(cq, &ucb->timeout_ev);
 
 	if (GNET_PROPERTY(udp_debug)) {
 		g_message("UDP [%s] RPC to %s timed out",
@@ -241,7 +241,7 @@ urpc_send(const char *what,
 				g_warning("unable to send whole %zu-byte UDP %s RPC to %s: "
 					"only sent %zu byte%s",
 					len, what, host_addr_port_to_string(addr, port),
-					r, 1 == r ? "" : "s");
+					r, plural(r));
 			}
 		}
 		socket_free_null(&s);
