@@ -85,6 +85,7 @@
 
 #include "core/gnet_stats.h"
 
+#include "lib/array_util.h"
 #include "lib/atoms.h"
 #include "lib/bstr.h"
 #include "lib/cq.h"
@@ -95,6 +96,7 @@
 #include "lib/hset.h"
 #include "lib/patricia.h"
 #include "lib/pmsg.h"
+#include "lib/pslist.h"
 #include "lib/str.h"
 #include "lib/stringify.h"
 #include "lib/walloc.h"
@@ -209,7 +211,7 @@ static double decimation_factor[KUID_RAW_BITSIZE];
 #define KEYS_DECIMATION_BASE 	1.2		/* Base for exponential decimation */
 #define KEYS_KEYDATA_VERSION	1		/* Serialization version number */
 
-static void keys_periodic_kball(cqueue_t *unused_cq, void *unused_obj);
+static void keys_periodic_kball(cqueue_t *cq, void *obj);
 
 /**
  * @return TRUE if key is stored here.
@@ -344,10 +346,12 @@ get_keydata(const kuid_t *id)
 
 	if (kd == NULL) {
 		if (dbmw_has_ioerr(db_keydata)) {
-			g_warning("DBMW \"%s\" I/O error, bad things could happen...",
+			s_warning_once_per(LOG_PERIOD_MINUTE,
+				"DBMW \"%s\" I/O error, bad things could happen...",
 				dbmw_name(db_keydata));
 		} else {
-			g_warning("key %s exists but was not found in DBMW \"%s\"",
+			s_warning_once_per(LOG_PERIOD_SECOND,
+				"key %s exists but was not found in DBMW \"%s\"",
 				kuid_to_hex_string(id), dbmw_name(db_keydata));
 		}
 		return NULL;
@@ -465,7 +469,7 @@ keys_expire_values(struct keyinfo *ki, time_t now)
 
 	if (kd->values != ki->values) {
 		str_bprintf(buf, sizeof buf, "expected %u value%s, has %u in keydata",
-			ki->values, 1 == ki->values ? "" : "s", kd->values);
+			ki->values, plural(ki->values), kd->values);
 		reason = buf;
 		goto discard_key;
 	}
@@ -498,7 +502,7 @@ keys_expire_values(struct keyinfo *ki, time_t now)
 
 	if (GNET_PROPERTY(dht_storage_debug) > 3)
 		g_debug("DHT STORE key %s has %d expired value%s out of %d",
-			kuid_to_hex_string(ki->kuid), expired, 1 == expired ? "" : "s",
+			kuid_to_hex_string(ki->kuid), expired, plural(expired),
 			ki->values);
 
 	if (next_expire != TIME_T_MAX)
@@ -519,7 +523,7 @@ discard_key:
 	 */
 
 	g_warning("DHT KEYS discarding corrupted key %s (%u value%s): %s",
-		kuid_to_hex_string(ki->kuid), ki->values, 1 == ki->values ? "" : "s",
+		kuid_to_hex_string(ki->kuid), ki->values, plural(ki->values),
 		reason);
 
 	keys_reclaim(ki, TRUE);
@@ -551,8 +555,7 @@ keys_get_status(const kuid_t *id, bool *full, bool *loaded)
 	if (GNET_PROPERTY(dht_storage_debug) > 1) {
 		g_debug("DHT STORE key %s holds %d/%d value%s, "
 			"load avg: get = %g [%s], store = %g [%s], expire in %s",
-			kuid_to_hex_string(id), ki->values, MAX_VALUES,
-			1 == ki->values ? "" : "s",
+			kuid_to_hex_string(id), ki->values, MAX_VALUES, plural(ki->values),
 			(int) (ki->get_req_load * 100) / 100.0,
 			ki->get_req_load >= LOAD_GET_THRESH ? "LOADED" : "OK",
 			(int) (ki->store_req_load * 100) / 100.0,
@@ -670,14 +673,9 @@ keys_remove_value(const kuid_t *id, const kuid_t *cid, uint64 dbkey)
 	g_assert(idx >= 0 && idx < kd->values);
 	g_assert(dbkey == kd->dbkeys[idx]);
 
-	if (idx < kd->values - 1) {
-		memmove(&kd->creators[idx], &kd->creators[idx+1],
-			sizeof(kd->creators[0]) * (kd->values - idx - 1));
-		memmove(&kd->dbkeys[idx], &kd->dbkeys[idx+1],
-			sizeof(kd->dbkeys[0]) * (kd->values - idx - 1));
-		memmove(&kd->expire[idx], &kd->expire[idx+1],
-			sizeof(kd->expire[0]) * (kd->values - idx - 1));
-	}
+	ARRAY_REMOVE(kd->creators, idx, kd->values);
+	ARRAY_REMOVE(kd->dbkeys,   idx, kd->values);
+	ARRAY_REMOVE(kd->expire,   idx, kd->values);
 
 	/*
 	 * We do not synchronously delete empty keys.
@@ -709,8 +707,7 @@ keys_remove_value(const kuid_t *id, const kuid_t *cid, uint64 dbkey)
 
 	if (GNET_PROPERTY(dht_storage_debug) > 2) {
 		g_debug("DHT STORE key %s now holds only %d/%d value%s, expire in %s",
-			kuid_to_hex_string(id), ki->values, MAX_VALUES,
-			1 == ki->values ? "" : "s",
+			kuid_to_hex_string(id), ki->values, MAX_VALUES, plural(ki->values),
 			compact_time(delta_time(ki->next_expire, tm_time())));
 	}
 }
@@ -757,7 +754,9 @@ keys_update_value(const kuid_t *id, const kuid_t *cid, time_t expire)
 			}
 		}
 
-		if (!found && GNET_PROPERTY(dht_keys_debug)) {
+		if (found) {
+			dbmw_write(db_keydata, id, kd, sizeof *kd);
+		} else if (GNET_PROPERTY(dht_keys_debug)) {
 			g_warning("DHT KEYS %s(): creator %s not found under %s",
 				G_STRFUNC, kuid_to_hex_string(cid), kuid_to_hex_string2(id));
 		}
@@ -818,7 +817,7 @@ keys_add_value(const kuid_t *id, const kuid_t *cid,
 		if (GNET_PROPERTY(dht_storage_debug) > 5)
 			g_debug("DHT STORE new %s %s (%zu common bit%s) with creator %s",
 				in_kball ? "key" : "cached key",
-				kuid_to_hex_string(id), common, 1 == common ? "" : "s",
+				kuid_to_hex_string(id), common, plural(common),
 				kuid_to_hex_string2(cid));
 
 		ki = allocate_keyinfo(id, common);
@@ -852,8 +851,7 @@ keys_add_value(const kuid_t *id, const kuid_t *cid,
 			g_debug("DHT STORE existing key %s (%u common bit%s) "
 				"has new creator %s",
 				kuid_to_hex_string(id), ki->common_bits,
-				1 == ki->common_bits ? "" : "s",
-				kuid_to_hex_string2(cid));
+				plural(ki->common_bits), kuid_to_hex_string2(cid));
 
 		/*
 		 * Keys are collected asynchronously, so it is possible that
@@ -887,20 +885,17 @@ keys_add_value(const kuid_t *id, const kuid_t *cid,
 				high = mid - 1;
 		}
 
-		/* Insert new item at `low' */
+		/* Make room for inserting new item at `low' */
 
-		if (low < kd->values) {
-			memmove(&kd->creators[low+1], &kd->creators[low],
-				sizeof(kd->creators[0]) * (kd->values - low));
-			memmove(&kd->dbkeys[low+1], &kd->dbkeys[low],
-				sizeof(kd->dbkeys[0]) * (kd->values - low));
-			memmove(&kd->expire[low+1], &kd->expire[low],
-				sizeof(kd->expire[0]) * (kd->values - low));
-		}
+		ARRAY_FIXED_MAKEROOM(kd->creators, low, kd->values);
+		ARRAY_FIXED_MAKEROOM(kd->dbkeys,   low, kd->values);
+		ARRAY_FIXED_MAKEROOM(kd->expire,   low, kd->values);
 
 		/* FALL THROUGH */
 
 	empty:
+		/* Insert new item at `low' */
+
 		kd->creators[low] = *cid;			/* struct copy */
 		kd->dbkeys[low] = dbkey;
 		kd->expire[low] = expire;
@@ -916,8 +911,7 @@ keys_add_value(const kuid_t *id, const kuid_t *cid,
 	if (GNET_PROPERTY(dht_storage_debug) > 2)
 		g_debug("DHT STORE %s key %s now holds %d/%d value%s",
 			&new_kd == kd ? "new" : "existing",
-			kuid_to_hex_string(id), ki->values, MAX_VALUES,
-			1 == ki->values ? "" : "s");
+			kuid_to_hex_string(id), ki->values, MAX_VALUES, plural(ki->values));
 }
 
 /**
@@ -1015,7 +1009,7 @@ keys_get(const kuid_t *id, dht_value_type_t type,
 			" with %d secondary key%s",
 			kuid_to_hex_string(id), ki->get_req_load, ki->get_requests,
 			dht_value_type_to_string(type),
-			secondary_count, 1 == secondary_count ? "" : "s");
+			secondary_count, plural(secondary_count));
 
 	*loadptr = ki->get_req_load;
 	ki->get_requests++;
@@ -1201,8 +1195,10 @@ keys_update_load(void *val, void *u)
 	 * Check for expired values.
 	 */
 
-	if (ctx->now >= ki->next_expire)
-		keys_expire_values(ki, ctx->now);
+	if (ctx->now >= ki->next_expire) {
+		if (!keys_expire_values(ki, ctx->now))
+			return TRUE;		/* ki was freed, delete entry */
+	}
 
 	/*
 	 * Collection of empty keys happens in a separate check because we also
@@ -1252,8 +1248,7 @@ keys_periodic_load(void *unused_obj)
 	if (GNET_PROPERTY(dht_storage_debug)) {
 		size_t keys_count = hikset_count(keys);
 		g_debug("DHT holding %zu value%s spread over %zu key%s",
-			ctx.values, 1 == ctx.values ? "" : "s",
-			keys_count, 1 == keys_count ? "" : "s");
+			ctx.values, plural(ctx.values), keys_count, plural(keys_count));
 	}
 
 	return TRUE;		/* Keep calling */
@@ -1316,13 +1311,11 @@ keys_update_kball(void)
 				kball.seeded ? "" : "(not seeded yet) ",
 				width == kball.width ? "remained at" :
 				width > kball.width ? "expanded to" : "shrunk to",
-				width, 1 == width ? "" : "s", kball.width);
+				width, plural(width), kball.width);
 			g_debug("DHT k-ball closest (%zu common bit%s) is %s",
-				cbits, 1 == cbits ? "" : "s",
-				knode_to_string(closest));
+				cbits, plural(cbits), knode_to_string(closest));
 			g_debug("DHT k-ball furthest (%zu common bit%s) is %s",
-				fbits, 1 == fbits ? "" : "s",
-				knode_to_string(furthest));
+				fbits, plural(fbits), knode_to_string(furthest));
 		}
 
 		STATIC_ASSERT(KUID_RAW_BITSIZE < 256);
@@ -1383,11 +1376,11 @@ keys_decimation_factor(const kuid_t *key)
  * Callout queue callback for k-ball updates.
  */
 static void
-keys_periodic_kball(cqueue_t *unused_cq, void *unused_obj)
+keys_periodic_kball(cqueue_t *cq, void *unused_obj)
 {
-	(void) unused_cq;
 	(void) unused_obj;
 
+	cq_zero(cq, &kball_ev);
 	install_periodic_kball(KBALL_PERIOD);
 	keys_update_kball();
 }
@@ -1538,7 +1531,7 @@ keys_init_keyinfo(void)
 	if (GNET_PROPERTY(dht_keys_debug)) {
 		size_t count = dbmw_count(db_keydata);
 		g_debug("DHT KEYS scanning %u persisted key%s",
-			(unsigned) count, 1 == count ? "" : "s");
+			(unsigned) count, plural(count));
 	}
 
 	ctx.our_kuid = get_our_kuid();
@@ -1549,7 +1542,7 @@ keys_init_keyinfo(void)
 	if (GNET_PROPERTY(dht_keys_debug)) {
 		size_t count = dbmw_count(db_keydata);
 		g_debug("DHT KEYS kept %u key%s, now loading associated values",
-			(unsigned) count, 1 == count ? "" : "s");
+			(unsigned) count, plural(count));
 	}
 
 	/*
@@ -1590,7 +1583,7 @@ keys_init_keyinfo(void)
 
 	if (GNET_PROPERTY(dht_keys_debug)) {
 		g_debug("DHT KEYS reloaded %zu key%s",
-			hikset_count(keys), 1 == hikset_count(keys) ? "" : "s");
+			hikset_count(keys), plural(hikset_count(keys)));
 	}
 
 	gnet_stats_set_general(GNR_DHT_KEYS_HELD, hikset_count(keys));
@@ -1651,7 +1644,7 @@ struct offload_context {
 	const kuid_t *our_kuid;			/**< Our KUID */
 	const kuid_t *remote_kuid;		/**< Remote node's KUID */
 	patricia_t *kclosest;			/**< Our k-closest alive nodes */
-	GSList *found;					/**< Target keys found */
+	pslist_t *found;				/**< Target keys found */
 	unsigned count;					/**< How many keys we found */
 };
 
@@ -1680,7 +1673,7 @@ keys_offload_prepare(void *val, void *data)
 	 */
 
 	if (patricia_closest(ctx->kclosest, id) == ctx->our_kuid) {
-		ctx->found = gm_slist_prepend_const(ctx->found, id);
+		ctx->found = pslist_prepend_const(ctx->found, id);
 		ctx->count++;
 	}
 }
@@ -1712,6 +1705,9 @@ keys_offload(const knode_t *kn)
 	if (kn->flags & (KNODE_F_FIREWALLED | KNODE_F_SHUTDOWNING))
 		return;
 
+	if G_UNLIKELY(NULL == keys)
+		return;							/* Key layer already shutdown */
+
 	if (
 		!dht_bootstrapped() ||			/* Not bootstrapped */
 		!keys_within_kball(kn->id) ||	/* Node KUID outside our k-ball */
@@ -1770,13 +1766,13 @@ keys_offload(const knode_t *kn)
 
 	if (debug) {
 		g_debug("DHT found %u/%zu offloading candidate%s",
-			ctx.count, hikset_count(keys), 1 == ctx.count ? "" : "s");
+			ctx.count, hikset_count(keys), plural(ctx.count));
 	}
 
 	if (ctx.count)
 		publish_offload(kn, ctx.found);
 
-	gm_slist_free_null(&ctx.found);
+	pslist_free_null(&ctx.found);
 }
 
 /**
