@@ -26,9 +26,11 @@
 
 #include "common.h"
 
-#include "lib/misc.h"
 #include "lib/hashlist.h"
+#include "lib/misc.h"
+#include "lib/pslist.h"
 #include "lib/vendors.h"
+
 #include "if/core/nodes.h"
 
 /***
@@ -48,6 +50,18 @@ enum {
 	SEARCH_F_PASSIVE 	= 1 << 0	/**< Start a passive ssearch */
 };
 
+/*
+ * Result sets `spam' flags.
+ */
+enum {
+	SPAM_F_ALT			= 1 << 5,		/** Carries alt-loc spam [UNUSED] */
+	SPAM_F_DUP			= 1 << 4,		/**< Duplicate entries in results */
+	SPAM_F_FAKE			= 1 << 3,		/**< Fake file */
+	SPAM_F_NAME			= 1 << 2,		/**< Carries filename spam */
+	SPAM_F_URL			= 1 << 1,		/**< Carries action URL spam */
+	SPAM_F_URN			= 1 << 0		/**< Carries spam known by URN */
+};
+
 /*
  * Result sets `status' flags.
  */
@@ -65,12 +79,12 @@ enum {
 	 ST_HOSTILE				= (1 << 19), /**< From an hostile host */
 	 ST_UNREQUESTED			= (1 << 18), /**< Unrequested (OOB) result */
 	 ST_EVIL				= (1 << 17), /**< Carries evil filename */
-	 ST_ALT_SPAM			= (1 << 16), /**< Carries alt-loc spam [UNUSED] */
-	 ST_DUP_SPAM			= (1 << 15), /**< Duplicate entries in results */
-	 ST_FAKE_SPAM			= (1 << 14), /**< Fake file */
-	 ST_NAME_SPAM			= (1 << 13), /**< Carries alt-loc spam */
-	 ST_URL_SPAM			= (1 << 12), /**< Carries action URL spam */
-	 ST_URN_SPAM			= (1 << 11), /**< Carries spam known by URN */
+	 ST_G2					= (1 << 16), /**< Sent by a G2 node */
+	 ST_UNUSED_4			= (1 << 15), /**< [UNUSED] */
+	 ST_UNUSED_3			= (1 << 14), /**< [UNUSED] */
+	 ST_UNUSED_2			= (1 << 13), /**< [UNUSED] */
+	 ST_UNUSED_1			= (1 << 12), /**< [UNUSED] */
+	 ST_SPAM				= (1 << 11), /**< Carries spam, flags in `spam' */
 	 ST_TLS					= (1 << 10), /**< Indicated support for TLS */
 	 ST_BH					= (1 << 9),	 /**< Browse Host support */
 	 ST_KNOWN_VENDOR		= (1 << 8),	 /**< Found known vendor code */
@@ -81,14 +95,7 @@ enum {
 	 ST_GGEP				= (1 << 3),	 /**< Trailer has a GGEP extension */
 	 ST_UPLOADED			= (1 << 2),	 /**< Is "stable", people downloaded */
 	 ST_BUSY				= (1 << 1),	 /**< Has currently no slots */
-	 ST_FIREWALL			= (1 << 0),	 /**< Is behind a firewall */
-
-	 ST_SPAM	= (	ST_ALT_SPAM
-			 		|ST_DUP_SPAM
-					|ST_FAKE_SPAM
-					|ST_NAME_SPAM
-					|ST_URL_SPAM
-					|ST_URN_SPAM)
+	 ST_FIREWALL			= (1 << 0)	 /**< Is behind a firewall */
 };
 
 /*
@@ -117,7 +124,7 @@ typedef struct gnet_results_set {
 	const char *version;		/**< Version information (atom) */
 	const char *query;			/**< Optional: Original query string (atom) */
 	gnet_host_vec_t *proxies;	/**< Optional: known push proxies */
-	GSList *records;
+	pslist_t *records;
 
 	time_t  stamp;				/**< Reception time of the hit */
 	vendor_code_t vcode;		/**< Vendor code */
@@ -131,12 +138,14 @@ typedef struct gnet_results_set {
 	uint8 hops;
 	uint8 ttl;
 	uint8 media;				/**< Optional: media type filtering */
+	uint8 spam;					/**< Spam flags */
 } gnet_results_set_t;
 
 /*
  * Result record flags
  */
 enum {
+	SR_ALLOC_NAME	= (1 << 11),	/* Set if filename was halloc()'ed */
 	SR_MEDIA		= (1 << 10),	/* Media type filter mismatch */
 	SR_PARTIAL_HIT	= (1 << 9),		/* Got a hit for a partial file */
 	SR_PUSH			= (1 << 8),		/* Servent firewalled, will need a PUSH */
@@ -186,7 +195,7 @@ typedef void (*search_request_listener_t) (
     query_type_t, const char *query, const host_addr_t addr, uint16);
 
 typedef void (*search_got_results_listener_t)
-    (GSList *, const struct guid *, const gnet_results_set_t *);
+    (pslist_t *, const struct guid *, const gnet_results_set_t *);
 
 typedef void (*search_status_change_listener_t)(gnet_search_t);
 
@@ -262,7 +271,7 @@ void search_request_listener_remove(search_request_listener_t);
 
 void search_associate_sha1(gnet_search_t sh, const struct sha1 *sha1);
 void search_dissociate_sha1(const struct sha1 *sha1);
-GSList *search_associated_sha1(gnet_search_t sh);
+pslist_t *search_associated_sha1(gnet_search_t sh);
 unsigned search_associated_sha1_count(gnet_search_t sh);
 
 const char *search_media_mask_to_string(unsigned mask);
