@@ -1,6 +1,4 @@
 /*
- * $Id: getdate.h,v 1.2 2005/06/25 01:37:42 daichik Exp $
- *
  * Copyright (c) 2002-2003, Raphael Manfredi
  *
  *----------------------------------------------------------------------
@@ -36,8 +34,7 @@
 #ifndef _getdate_h_
 #define _getdate_h_
 
-#include <time.h>
-#include <sys/time.h>
+#include "common.h"
 
 time_t date2time(const char *p, time_t now);
 
