@@ -102,13 +102,35 @@ gnet_host_hash2(const void *key)
 	return host_addr_hash2(addr) ^ port_hash2(port);
 }
 
+/**
+ * Compare function which returns TRUE if the hosts are equivalent.
+ *
+ * @note
+ * The host addresses need not be equal if the conversion to an IPv4 address
+ * makes one of them equivalent to the other.
+ *
+ * @attention
+ * This routine compares the addreses in a way that makes it unsuitable for
+ * hash tables or sets because we're not testing for true equality, rather
+ * for equivalence modulo an IPv6 to IPv4 conversion.  Therefore, it breaks the
+ * implicit assumption that equal items will hash to the same value!
+ */
+bool
+gnet_host_equiv(const void *v1, const void *v2)
+{
+	const gnet_host_t *h1 = v1, *h2 = v2;
+
+	return gnet_host_get_port(h1) == gnet_host_get_port(h2) &&
+		host_addr_equiv(gnet_host_get_addr(h1), gnet_host_get_addr(h2));
+}
+
 /**
  * Compare function which returns TRUE if the hosts are equal.
  *
  * @note For use in hash tables and sets.
  */
-G_GNUC_HOT bool
-gnet_host_eq(const void *v1, const void *v2)
+bool
+gnet_host_equal(const void *v1, const void *v2)
 {
 	const gnet_host_t *h1 = v1, *h2 = v2;
 
@@ -117,25 +139,14 @@ gnet_host_eq(const void *v1, const void *v2)
 }
 
 /**
- * Compare function which returns 0 if the hosts are equal, otherwise 1.
- *
- * @note For use in g_list_find_custom.
- */
-int
-gnet_host_cmp(const void *v1, const void *v2)
-{
-	return gnet_host_eq(v1, v2) ? 0 : 1;
-}
-
-/**
- * Compare function which returns TRUE if the host addresses are equal.
+ * Compare function which returns TRUE if the host addresses are equivalent.
  */
 bool
-gnet_host_addr_eq(const void *v1, const void *v2)
+gnet_host_addr_equiv(const void *v1, const void *v2)
 {
 	const gnet_host_t *h1 = v1, *h2 = v2;
 
-	return host_addr_equal(gnet_host_get_addr(h1), gnet_host_get_addr(h2));
+	return host_addr_equiv(gnet_host_get_addr(h1), gnet_host_get_addr(h2));
 }
 
 /**
@@ -366,7 +377,7 @@ gnet_host_vec_to_string(const gnet_host_vec_t *hvec)
 		gchar buf[128];
 
 		if (i > 0) {
-			str_cat(s, ", ");
+			STR_CAT(s, ", ");
 		}
 		host = gnet_host_vec_get(hvec, i);
 		host_addr_port_to_string_buf(gnet_host_get_addr(&host),
@@ -579,6 +590,17 @@ gnet_host_vec_from_vector(vector_t *vec)
 	return gnet_host_vec_from_sequence(sequence_fill_from_vector(&seq, vec));
 }
 
+/**
+ * Create a new Gnutella host vector out of a pslist_t of gnet_host_t items.
+ */
+gnet_host_vec_t *
+gnet_host_vec_from_pslist(pslist_t *pl)
+{
+	sequence_t seq;
+
+	return gnet_host_vec_from_sequence(sequence_fill_from_pslist(&seq, pl));
+}
+
 /**
  * Create a new Gnutella host vector out of a GSList of gnet_host_t items.
  */
