@@ -50,7 +50,8 @@
 #include "common.h"
 
 #include "fs_free_space.h"
-#include "tm.h"
+#include "pslist.h"
+#include "sha1.h"
 #include "vmm.h"
 
 #define SIZE_FIELD_MAX		64	/**< Max size of sprintf-ed size quantity */
@@ -142,9 +143,6 @@ short_string_t short_rate_get_string(uint64 rate, bool metric);
 /*
  * SHA1<->base32 string conversion
  */
-typedef struct sha1 {
-	char data[SHA1_RAW_SIZE];
-} sha1_t;
 
 #define SHA1_URN_LENGTH	(CONST_STRLEN("urn:sha1:") + SHA1_BASE32_SIZE)
 
@@ -195,7 +193,7 @@ const char *guid_to_string(const struct guid *);
  * GUID<->base32 string conversion
  */
 const char *guid_base32_str(const struct guid *);
-const struct guid *base32_to_guid(const char *);
+const struct guid *base32_to_guid(const char *, struct guid *);
 
 /*
  * Generic binary to hexadecimal conversion.
@@ -273,6 +271,8 @@ float force_range(float value, float min, float max);
 const char *short_filename(const char *fullname);
 char *data_hex_str(const char *data, size_t len);
 char *xml_indent(const char *text);
+pslist_t *dirlist_parse(const char *dirs);
+char *dirlist_to_string(const pslist_t *pl_dirs);
 
 #if defined(S_IROTH) && defined(S_IXOTH)
 /* 0755 */
@@ -296,6 +296,16 @@ void normalize_dir_separators(char *);
 size_t memcmp_diff(const void *a, const void *b, size_t n);
 int bitcmp(const void *s1, const void *s2, size_t n);
 
+/**
+ * Returns the length of the string plus one, i.o.w.
+ * the required buffer size in bytes.
+ */
+static inline size_t
+strsize(const char *src)
+{
+	return strlen(src) + 1;
+}
+
 /**
  * An strcpy() that returns the length of the copied string.
  */
@@ -383,6 +393,22 @@ clamp_memset(void *dst, size_t dst_size, char c, size_t n)
 	return n;
 }
 
+/**
+ * Compare at most MIN(a_len, b_len) bytes between "a" and "b".
+ *
+ * @param a		the first buffer.
+ * @param a_len the length of the first buffer, in bytes
+ * @param b		the second buffer.
+ * @param b_len the length of the second buffer, in bytes
+ *
+ * @return the result of the memcmp() operation (-1, 0, +1).
+ */
+static inline int
+clamp_memcmp(const void *a, size_t a_len, const void *b, size_t b_len)
+{
+	return memcmp(a, b, MIN(a_len, b_len));
+}
+
 /**
  * Copies at most MIN(dst_size - 1, src_len) characters from the buffer "src"
  * to the buffer "dst", ensuring the resulting string in "dst" is
