@@ -41,43 +41,66 @@
 
 #ifndef _SHA_enum_
 #define _SHA_enum_
-enum
+enum SHA_code
 {
-    shaSuccess = 0,
-    shaNull,            /**< Null pointer parameter */
-    shaInputTooLong,    /**< input data too long */
-    shaStateError       /**< called Input after Result */
+    SHA_SUCCESS = 0,       /**< OK */
+    SHA_NULL,              /**< Null pointer parameter */
+    SHA_INPUT_TOO_LONG,    /**< input data too long */
+    SHA_STATE_ERROR        /**< called Input after Result */
 };
 #endif
 
-struct sha1;
+typedef struct sha1 {
+	char data[SHA1_RAW_SIZE];
+} sha1_t;
+
+enum SHA1_context_magic { SHA1_CONTEXT_MAGIC = 0x73accbce };
 
 /**
  *  This structure will hold context information for the SHA-1
  *  hashing operation
  */
-typedef struct SHA1Context
+typedef struct SHA1_context {
+	enum SHA1_context_magic magic;		/* Magic number */
+	uint32 ihash[SHA1_RAW_SIZE / 4];	/* Intermediate Message Digest  */
+	uint64 length;            /* Message length in bits */
+	int midx;                 /* Index into message block array */
+	uint8 mblock[64];         /* 512-bit message blocks */
+	bool computed;            /* Is the digest computed? */
+	enum SHA_code corrupted;  /* Is the message digest corrupted? */
+} SHA1_context;
+
+static inline void
+SHA1_check(const SHA1_context * const ctx)
 {
-    uint32 Intermediate_Hash[SHA1_RAW_SIZE / 4]; /* Message Digest  */
-
-    uint64 Length;            /* Message length in bits      */
-
-                              /* Index into message block array   */
-    int Message_Block_Index;
-    uint8 Message_Block[64];  /* 512-bit message blocks      */
-
-    int Computed;             /* Is the digest computed?         */
-    int Corrupted;            /* Is the message digest corrupted? */
-} SHA1Context;
+	g_assert(NULL == ctx || SHA1_CONTEXT_MAGIC == ctx->magic);
+}
 
 /*
  *  Function Prototypes
  */
 
+int SHA1_reset(SHA1_context *);
+int SHA1_input(SHA1_context *, const void *, size_t);
+int SHA1_result(SHA1_context *, struct sha1 *digest);
 
-int SHA1Reset(  SHA1Context *);
-int SHA1Input(  SHA1Context *, const void *, size_t);
-int SHA1Result( SHA1Context *, struct sha1 *Message_Digest);
+/**
+ * Feed the SHA1 context with the content of a variable.
+ */
+#define SHA1_INPUT(c,v)		SHA1_input((c), &(v), sizeof(v))
+
+/**
+ * Compute the SHA1 digest of (structure) ``x'' into ``d''.
+ *
+ * ``x'' is the data structure, and ``sizeof(x)'' gives the size to hash.
+ * ``d'' is the address of the output digest (sha1_t *).
+ */
+#define SHA1_COMPUTE(x,d) G_STMT_START {	\
+	SHA1_context c_;						\
+	SHA1_reset(&c_);						\
+	SHA1_input(&c_, &(x), sizeof(x));		\
+	SHA1_result(&c_, (d));					\
+} G_STMT_END
 
 #endif /* _sha1_h_ */
 
