@@ -23,10 +23,15 @@
 #include "lib/halloc.h"
 #include "lib/log.h"
 #include "lib/misc.h"
+#include "lib/mutex.h"
 #include "lib/pow2.h"
 #include "lib/random.h"
 #include "lib/str.h"
+#include "lib/stringify.h"
+#include "lib/thread.h"
+#include "lib/vmm.h"
 #include "lib/walloc.h"
+#include "lib/xmalloc.h"
 
 #include "lib/override.h"		/* Must be the last header included */
 
@@ -42,6 +47,63 @@ static datum getnext(DBM *);
 static bool makroom(DBM *, long, size_t);
 static void validpage(DBM *, long);
 
+/*
+ * Thread-safety macros.
+ */
+
+#ifdef THREADS
+
+#define sdbm_synchronize(s) G_STMT_START {		\
+	if G_UNLIKELY((s)->lock != NULL) { 			\
+		DBM *ws = deconstify_pointer(s);	\
+		mutex_lock(ws->lock);					\
+	}											\
+} G_STMT_END
+
+#define sdbm_unsynchronize(s) G_STMT_START {	\
+	if G_UNLIKELY((s)->lock != NULL) { 			\
+		DBM *ws = deconstify_pointer(s);	\
+		mutex_unlock(ws->lock);					\
+	}											\
+} G_STMT_END
+
+#define sdbm_return(s, v) G_STMT_START {		\
+	if G_UNLIKELY((s)->lock != NULL) 			\
+		mutex_unlock((s)->lock);				\
+	return v;									\
+} G_STMT_END
+
+#define sdbm_return_datum(s, v) G_STMT_START {	\
+	datum *rv = &(v);							\
+	if G_UNLIKELY((s)->lock != NULL) { 			\
+		rv = sdbm_thread_datum((s), &(v));		\
+		mutex_unlock((s)->lock);				\
+	}											\
+	return *rv;									\
+} G_STMT_END
+
+#define sdbm_return_void(s) G_STMT_START {		\
+	if G_UNLIKELY((s)->lock != NULL) 			\
+		mutex_unlock((s)->lock);				\
+	return;										\
+} G_STMT_END
+
+#define assert_sdbm_locked(s) G_STMT_START {	\
+	if G_UNLIKELY((s)->lock != NULL) 			\
+		assert_mutex_is_owned((s)->lock);		\
+} G_STMT_END
+
+#else	/* !THREADS */
+
+#define sdbm_synchronize(s)
+#define sdbm_unsynchronize(s)
+#define sdbm_return(s, v)			return v
+#define sdbm_return_datum(s, v)		return v
+#define sdbm_return_void(s)			return
+#define assert_sdbm_locked(s)
+
+#endif	/* THREADS */
+
 static inline int
 bad(const datum item)
 {
@@ -218,6 +280,11 @@ sdbm_alloc(void)
 	db->magic = SDBM_MAGIC;
 	db->pagf = -1;
 	db->dirf = -1;
+
+#ifdef THREADS
+	db->iterid = THREAD_INVALID_ID;
+#endif
+
 	return db;
 }
 
@@ -225,6 +292,22 @@ static void
 sdbm_free(DBM *db)
 {
 	sdbm_check(db);
+
+#ifdef THREADS
+	if (db->returned != NULL) {
+		int i;
+
+		for (i = 0; i < THREAD_MAX; i++) {
+			datum *r = &db->returned[i];
+
+			if (r->dsize != 0)
+				XFREE_NULL(r->dptr);	/* Free thread-private data copy */
+		}
+
+		XFREE_NULL(db->returned);
+	}
+#endif
+
 	db->magic = 0;
 	WFREE(db);
 }
@@ -368,7 +451,7 @@ success:
 
 		if ((flags & (O_RDWR | O_WRONLY) && (flags & O_TRUNC))) {
 			if (-1 == unlink(datname) && ENOENT != errno)
-				g_warning("%s(): cannot delete \"%s\": %m", G_STRFUNC, datname);
+				s_warning("%s(): cannot delete \"%s\": %m", G_STRFUNC, datname);
 		}
 	}
 #else
@@ -383,24 +466,126 @@ success:
 	return db;
 }
 
+#ifdef THREADS
+/**
+ * Mark newly created database as being thread-safe.
+ *
+ * This will make all external operations on the database thread-safe.
+ */
+void
+sdbm_thread_safe(DBM *db)
+{
+	sdbm_check(db);
+	g_assert(NULL == db->lock);
+
+	WALLOC0(db->lock);
+	mutex_init(db->lock);
+	XMALLOC0_ARRAY(db->returned, THREAD_MAX);
+}
+
+/**
+ * Lock the database to allow a sequence of operations to be atomically
+ * conducted.
+ *
+ * It is possible to lock the database several times as long as each locking
+ * is paired with a corresponding unlocking in the execution flow.
+ *
+ * The database must have been marked thread-safe already.
+ */
+void
+sdbm_lock(DBM *db)
+{
+	sdbm_check(db);
+	g_assert_log(db->lock != NULL,
+		"%s(): SDBM \"%s\" not marked thread-safe", G_STRFUNC, sdbm_name(db));
+
+	mutex_lock(db->lock);
+}
+
+/*
+ * Release lock on database.
+ *
+ * The database must have been marked thread-safe already and locked by the
+ * calling thread.
+ */
+void
+sdbm_unlock(DBM *db)
+{
+	sdbm_check(db);
+	g_assert_log(db->lock != NULL,
+		"%s(): SDBM \"%s\" not marked thread-safe", G_STRFUNC, sdbm_name(db));
+
+	mutex_unlock(db->lock);
+}
+
+/**
+ * Allocate a thread-private datum to be returned to the thread.
+ *
+ * Although the value returned is thread-private, its lifespan is limited to
+ * the next call made to the SDBM API by the thread.
+ *
+ * A thread-private datum contains a copy of the data returned to the thread,
+ * since we cannot point into the internal SDBM data structures like the LRU
+ * page cache: any concurrent access could make the data stale.
+ *
+ * @param v		a pointer to the datum returned by the function
+ *
+ * @return a pointer to the thread-private datum with data from ``d'' copied.
+ */
+static datum *
+sdbm_thread_datum(DBM *db, datum *v)
+{
+	datum *r;
+	uint stid = thread_small_id();
+
+	sdbm_check(db);
+	g_assert(stid < THREAD_MAX);
+	g_assert(db->returned != NULL);
+
+	r = &db->returned[stid];
+
+	if (r->dsize != 0)
+		xfree(r->dptr);						/* Free old value */
+
+	if (v->dsize != 0) {
+		r->dptr = xcopy(v->dptr, v->dsize);	/* Copy value to be returned */
+		r->dsize = v->dsize;
+	} else if (NULL == v->dptr) {
+		*r = nullitem;
+	} else {
+		/*
+		 * It's possible to have a zero-sized value stored, and if we come
+		 * here then v->dsize = 0 and v->dptr != NULL.
+		 *
+		 * To prevent any reference to the pointer, we use the VMM trap page.
+		 */
+
+		r->dptr = deconstify_pointer(vmm_trap_page());
+		r->dsize = 0;
+	}
+
+	return r;
+}
+#endif	/* THREADS */
+
 static void
 log_sdbmstats(DBM *db)
 {
 	sdbm_check(db);
 
-	g_info("sdbm: \"%s\" page reads = %lu, page writes = %lu (forced %lu)",
+	s_info("sdbm: \"%s\" page reads = %lu, page writes = %lu (forced %lu)",
 		sdbm_name(db), db->pagread, db->pagwrite, db->pagwforced);
-	g_info("sdbm: \"%s\" dir reads = %lu, dir writes = %lu (deferred %lu)",
+	s_info("sdbm: \"%s\" dir reads = %lu, dir writes = %lu (deferred %lu)",
 		sdbm_name(db), db->dirread, db->dirwrite, db->dirwdelayed);
-	g_info("sdbm: \"%s\" page blocknum hits = %.2f%% on %lu request%s",
+	s_info("sdbm: \"%s\" page blocknum hits = %.2f%% on %lu request%s",
 		sdbm_name(db), db->pagbno_hit * 100.0 / MAX(db->pagfetch, 1),
-		db->pagfetch, 1 == db->pagfetch ? "" : "s");
-	g_info("sdbm: \"%s\" dir blocknum hits = %.2f%% on %lu request%s",
+		db->pagfetch, plural(db->pagfetch));
+	s_info("sdbm: \"%s\" dir blocknum hits = %.2f%% on %lu request%s",
 		sdbm_name(db), db->dirbno_hit * 100.0 / MAX(db->dirfetch, 1),
-		db->dirfetch, 1 == db->dirfetch ? "" : "s");
-	g_info("sdbm: \"%s\" inplace value writes = %.2f%% on %lu occurence%s",
+		db->dirfetch, plural(db->dirfetch));
+	s_info("sdbm: \"%s\" inplace value writes = %.2f%% on %lu occurence%s",
 		sdbm_name(db), db->repl_inplace * 100.0 / MAX(db->repl_stores, 1),
-		db->repl_stores, 1 == db->repl_stores ? "" : "s");
+		db->repl_stores, plural(db->repl_stores));
 }
 
 static void
@@ -409,32 +594,32 @@ log_sdbm_warnings(DBM *db)
 	sdbm_check(db);
 
 	if (db->flags & DBM_BROKEN) {
-		g_warning("sdbm: \"%s\" descriptor was broken by failed renaming",
+		s_warning("sdbm: \"%s\" descriptor was broken by failed renaming",
 			sdbm_name(db));
 	}
 	if (db->bad_pages) {
-		g_warning("sdbm: \"%s\" read %lu corrupted page%s (zero-ed on the fly)",
-			sdbm_name(db), db->bad_pages, 1 == db->bad_pages ? "" : "s");
+		s_warning("sdbm: \"%s\" read %lu corrupted page%s (zero-ed on the fly)",
+			sdbm_name(db), db->bad_pages, plural(db->bad_pages));
 	}
 	if (db->removed_keys) {
-		g_warning("sdbm: \"%s\" removed %lu key%s not belonging to their page",
-			sdbm_name(db), db->removed_keys, 1 == db->removed_keys ? "" : "s");
+		s_warning("sdbm: \"%s\" removed %lu key%s not belonging to their page",
+			sdbm_name(db), db->removed_keys, plural(db->removed_keys));
 	}
 	if (db->read_errors || db->write_errors) {
-		g_warning("sdbm: \"%s\" "
+		s_warning("sdbm: \"%s\" "
 			"ERRORS: read = %lu, write = %lu (%lu in flushes, %lu in splits)",
 			sdbm_name(db),
 			db->read_errors, db->write_errors,
 			db->flush_errors, db->spl_errors);
 	}
 	if (db->spl_corrupt) {
-		g_warning("sdbm: \"%s\" %lu failed page split%s could not be undone",
-			sdbm_name(db), db->spl_corrupt, 1 == db->spl_corrupt ? "" : "s");
+		s_warning("sdbm: \"%s\" %lu failed page split%s could not be undone",
+			sdbm_name(db), db->spl_corrupt, plural(db->spl_corrupt));
 	}
 #ifdef BIGDATA
 	if (db->bad_bigkeys) {
-		g_warning("sdbm: \"%s\" encountered %lu bad big key%s",
-			sdbm_name(db), db->bad_bigkeys, 1 == db->bad_bigkeys ? "" : "s");
+		s_warning("sdbm: \"%s\" encountered %lu bad big key%s",
+			sdbm_name(db), db->bad_bigkeys, plural(db->bad_bigkeys));
 	}
 #endif
 }
@@ -448,6 +633,8 @@ log_sdbm_warnings(DBM *db)
 static bool
 fetch_pagbuf(DBM *db, long pagnum)
 {
+	assert_sdbm_locked(db);
+
 	db->pagfetch++;
 
 #ifdef LRU
@@ -491,7 +678,7 @@ fetch_pagbuf(DBM *db, long pagnum)
 		db->pagread++;
 		got = compat_pread(db->pagf, db->pagbuf, DBM_PBLKSIZ, OFF_PAG(pagnum));
 		if G_UNLIKELY(got < 0) {
-			g_critical("sdbm: \"%s\": cannot read page #%ld: %m",
+			s_critical("sdbm: \"%s\": cannot read page #%ld: %m",
 				sdbm_name(db), pagnum);
 			ioerr(db, FALSE);
 			db->pagbno = -1;
@@ -499,12 +686,12 @@ fetch_pagbuf(DBM *db, long pagnum)
 		}
 		if G_UNLIKELY(got < DBM_PBLKSIZ) {
 			if (got > 0)
-				g_critical("sdbm: \"%s\": partial read (%u bytes) of page #%ld",
+				s_critical("sdbm: \"%s\": partial read (%u bytes) of page #%ld",
 					sdbm_name(db), (unsigned) got, pagnum);
 			memset(db->pagbuf + got, 0, DBM_PBLKSIZ - got);
 		}
 		if G_UNLIKELY(!sdbm_internal_chkpage(db->pagbuf)) {
-			g_critical("sdbm: \"%s\": corrupted page #%ld, clearing",
+			s_critical("sdbm: \"%s\": corrupted page #%ld, clearing",
 				sdbm_name(db), pagnum);
 			memset(db->pagbuf, 0, DBM_PBLKSIZ);
 			db->bad_pages++;
@@ -526,6 +713,8 @@ fetch_pagbuf(DBM *db, long pagnum)
 static bool
 flush_pagbuf(DBM *db)
 {
+	assert_sdbm_locked(db);
+
 #ifdef LRU
 	return dirtypag(db, FALSE);	/* Current (cached) page buffer is dirty */
 #else
@@ -541,6 +730,8 @@ flush_pagbuf(DBM *db)
 static bool
 force_flush_pagbuf(DBM *db, bool force)
 {
+	assert_sdbm_locked(db);
+
 	if (force)
 		db->pagwforced++;
 	return dirtypag(db, force);	/* Current (cached) page buffer is dirty */
@@ -556,6 +747,8 @@ flush_dirbuf(DBM *db)
 {
 	ssize_t w;
 
+	assert_sdbm_locked(db);
+
 	db->dirwrite++;
 	w = compat_pwrite(db->dirf, db->dirbuf, DBM_DBLKSIZ, OFF_DIR(db->dirbno));
 
@@ -573,7 +766,7 @@ flush_dirbuf(DBM *db)
 #endif
 
 	if G_UNLIKELY(w != DBM_DBLKSIZ) {
-		g_critical("sdbm: \"%s\": cannot flush dir block #%ld: %s",
+		s_critical("sdbm: \"%s\": cannot flush dir block #%ld: %s",
 			sdbm_name(db), db->dirbno,
 			-1 == w ? g_strerror(errno) : "partial write");
 
@@ -590,7 +783,7 @@ sdbm_unlink_file(const char *name, const char *path)
 	g_assert(path != NULL);
 
 	if (-1 == unlink(path))
-		g_critical("sdbm: \"%s\": cannot unlink \"%s\": %m", name, path);
+		s_critical("sdbm: \"%s\": cannot unlink \"%s\": %m", name, path);
 }
 
 /**
@@ -604,10 +797,11 @@ static void
 sdbm_close_internal(DBM *db, bool clearfiles, bool destroy)
 {
 	sdbm_check(db);
+	assert_sdbm_locked(db);
 
 #ifdef LRU
 	if (!clearfiles && db->dirbuf_dirty && !(db->flags & DBM_BROKEN))
-		flush_dirbuf(db);
+		(void) flush_dirbuf(db);
 	lru_close(db);
 #else
 	WFREE_NULL(db->pagbuf, DBM_PBLKSIZ);
@@ -639,8 +833,13 @@ sdbm_close_internal(DBM *db, bool clearfiles, bool destroy)
 	HFREE_NULL(db->datname);
 #endif
 
-	if (destroy)
+	if (destroy) {
+		if (db->lock != NULL) {
+			mutex_destroy(db->lock);
+			WFREE(db->lock);
+		}
 		sdbm_free(db);
+	}
 }
 
 /**
@@ -652,6 +851,7 @@ sdbm_unlink(DBM *db)
 	if G_UNLIKELY(db == NULL)
 		return;
 
+	sdbm_synchronize(db);
 	sdbm_close_internal(db, TRUE, TRUE);
 }
 
@@ -670,6 +870,8 @@ sdbm_close(DBM *db)
 
 	sdbm_check(db);
 
+	sdbm_synchronize(db);
+
 #ifdef LRU
 	clearfiles = db->is_volatile;
 #else
@@ -695,16 +897,24 @@ sdbm_fetch(DBM *db, datum key)
 		return nullitem;
 	}
 	sdbm_check(db);
+
+	sdbm_synchronize(db);
+
 	if G_UNLIKELY(db->flags & DBM_BROKEN) {
 		errno = ESTALE;
-		return nullitem;
+		goto null;
 	}
+
 	SDBM_WARN_ITERATING(db);
-	if (getpage(db, exhash(key)))
-		return getpair(db, db->pagbuf, key);
+
+	if (getpage(db, exhash(key))) {
+		datum value = getpair(db, db->pagbuf, key);
+		sdbm_return_datum(db, value);
+	}
 
 	ioerr(db, FALSE);
-	return nullitem;
+null:
+	sdbm_return(db, nullitem);
 }
 
 /**
@@ -720,16 +930,22 @@ sdbm_exists(DBM *db, datum key)
 		return -1;
 	}
 	sdbm_check(db);
+
+	sdbm_synchronize(db);
+
 	if G_UNLIKELY(db->flags & DBM_BROKEN) {
 		errno = ESTALE;
-		return -1;
+		goto error;
 	}
 	SDBM_WARN_ITERATING(db);
-	if (getpage(db, exhash(key)))
-		return exipair(db, db->pagbuf, key);
+	if (getpage(db, exhash(key))) {
+		int exists = exipair(db, db->pagbuf, key);
+		sdbm_return(db, exists);
+	}
 
 	ioerr(db, FALSE);
-	return -1;
+error:
+	sdbm_return(db, -1);
 }
 
 /**
@@ -740,31 +956,36 @@ sdbm_exists(DBM *db, datum key)
 int
 sdbm_delete(DBM *db, datum key)
 {
+	int status = -1;
+
 	if G_UNLIKELY(db == NULL || bad(key)) {
 		errno = EINVAL;
 		return -1;
 	}
 	sdbm_check(db);
+
+	sdbm_synchronize(db);
+
 	if G_UNLIKELY(db->flags & DBM_RDONLY) {
 		errno = EPERM;
-		return -1;
+		goto done;
 	}
 	if G_UNLIKELY(db->flags & DBM_IOERR_W) {
 		errno = EIO;
-		return -1;
+		goto done;
 	}
 	if G_UNLIKELY(db->flags & DBM_BROKEN) {
 		errno = ESTALE;
-		return -1;
+		goto done;
 	}
 	SDBM_WARN_ITERATING(db);
 	if G_UNLIKELY(!getpage(db, exhash(key))) {
 		ioerr(db, FALSE);
-		return -1;
+		goto done;
 	}
 	if (!delpair(db, db->pagbuf, key)) {
 		errno = 0;
-		return -1;
+		goto done;
 	}
 
 	/*
@@ -772,9 +993,12 @@ sdbm_delete(DBM *db, datum key)
 	 */
 
 	if G_UNLIKELY(!flush_pagbuf(db))
-		return -1;
+		goto done;
 
-	return 0;
+	status = 0;
+
+done:
+	sdbm_return(db, status);
 }
 
 /**
@@ -797,6 +1021,8 @@ storepair(DBM *db, datum key, datum val, int flags, bool *existed)
 	bool need_split = FALSE;
 	int result = 0;
 
+	assert_sdbm_locked(db);
+
 	if G_UNLIKELY(0 == val.dsize) {
 		val.dptr = "";
 	}
@@ -924,9 +1150,16 @@ inserted:
 int
 sdbm_store(DBM *db, datum key, datum val, int flags)
 {
+	int r;
+
 	sdbm_check(db);
+
+	sdbm_synchronize(db);
+
 	SDBM_WARN_ITERATING(db);
-	return storepair(db, key, val, flags, NULL);
+	r = storepair(db, key, val, flags, NULL);
+
+	sdbm_return(db, r);
 }
 
 /**
@@ -941,9 +1174,16 @@ sdbm_store(DBM *db, datum key, datum val, int flags)
 int
 sdbm_replace(DBM *db, datum key, datum val, bool *existed)
 {
+	int r;
+
 	sdbm_check(db);
+
+	sdbm_synchronize(db);
+
 	SDBM_WARN_ITERATING(db);
-	return storepair(db, key, val, DBM_REPLACE, existed);
+	r = storepair(db, key, val, DBM_REPLACE, existed);
+
+	sdbm_return(db, r);
 }
 
 /*
@@ -962,6 +1202,8 @@ makroom(DBM *db, long int hash, size_t need)
 	char *New = (char *) twin;
 	int smax = DBM_SPLTMAX;
 
+	assert_sdbm_locked(db);
+
 	do {
 		bool fits;		/* Can we fit new pair in the split page? */
 
@@ -1100,7 +1342,7 @@ makroom(DBM *db, long int hash, size_t need)
 			db->pagwrite++,
 			compat_pwrite(db->pagf, New, DBM_PBLKSIZ, OFF_PAG(newp)) < 0)
 		) {
-			g_warning("sdbm: \"%s\": cannot flush new page #%ld: %m",
+			s_warning("sdbm: \"%s\": cannot flush new page #%ld: %m",
 				sdbm_name(db), newp);
 			ioerr(db, TRUE);
 			memcpy(pag, cur, DBM_PBLKSIZ);	/* Undo split */
@@ -1157,7 +1399,7 @@ makroom(DBM *db, long int hash, size_t need)
 		 */
 
 		if G_UNLIKELY(!setdbit(db, db->curbit)) {
-			g_critical("sdbm: \"%s\": "
+			s_critical("sdbm: \"%s\": "
 				"cannot set bit in forest bitmap for 0x%lx",
 				sdbm_name(db), db->curbit);
 			db->spl_errors++;
@@ -1183,7 +1425,7 @@ makroom(DBM *db, long int hash, size_t need)
 	 * we still cannot fit the key. say goodnight.
 	 */
 
-	g_critical("sdbm: \"%s\": cannot insert after DBM_SPLTMAX (%d) attempts",
+	s_critical("sdbm: \"%s\": cannot insert after DBM_SPLTMAX (%d) attempts",
 		sdbm_name(db), DBM_SPLTMAX);
 
 	return FALSE;
@@ -1232,7 +1474,7 @@ restore:
 		if (failed) {
 			db->spl_errors++;
 			db->spl_corrupt++;
-			g_critical("sdbm: \"%s\": cannot undo split of page #%lu: %m",
+			s_critical("sdbm: \"%s\": cannot undo split of page #%lu: %m",
 				sdbm_name(db), curbno);
 		}
 	} else {
@@ -1249,7 +1491,7 @@ restore:
 #endif
 		memset(New, 0, DBM_PBLKSIZ);
 		if (compat_pwrite(db->pagf, New, DBM_PBLKSIZ, OFF_PAG(newp)) < 0) {
-			g_critical("sdbm: \"%s\": cannot zero-back new split page #%ld: %m",
+			s_critical("sdbm: \"%s\": cannot zero-back new split page #%ld: %m",
 				sdbm_name(db), newp);
 			ioerr(db, TRUE);
 			db->spl_errors++;
@@ -1262,7 +1504,7 @@ restore:
 	/* FALL THROUGH */
 
 aborted:
-	g_warning("sdbm: \"%s\": aborted page split operation", sdbm_name(db));
+	s_warning("sdbm: \"%s\": aborted page split operation", sdbm_name(db));
 	return FALSE;
 }
 
@@ -1270,19 +1512,25 @@ static datum
 iteration_done(DBM *db, bool completed)
 {
 	g_assert(db != NULL);
+	assert_sdbm_locked(db);
 
 #ifdef BIGDATA
 	if (db->flags & DBM_KEYCHECK) {
 		size_t adj = big_check_end(db, completed);
 
 		if (adj != 0) {
-			g_warning("sdbm: \"%s\": database may have lost entries",
+			s_warning("sdbm: \"%s\": database may have lost entries",
 				sdbm_name(db));
 		}
 	}
 #endif
 
+#ifdef THREADS
+	db->iterid = THREAD_INVALID_ID;
+#endif
+
 	db->flags &= ~(DBM_KEYCHECK | DBM_ITERATING);	/* Iteration done */
+
 	return nullitem;
 }
 
@@ -1294,22 +1542,45 @@ iteration_done(DBM *db, bool completed)
 datum
 sdbm_firstkey(DBM *db)
 {
+	datum value;
+
 	if G_UNLIKELY(db == NULL) {
 		errno = EINVAL;
-		return iteration_done(db, FALSE);
+		return nullitem;
 	}
 	sdbm_check(db);
 
+	sdbm_synchronize(db);
+
 	if G_UNLIKELY(db->flags & DBM_BROKEN) {
 		errno = ESTALE;
-		return nullitem;
+		value = nullitem;
+		goto done;
 	}
 
 	if G_UNLIKELY(db->flags & DBM_ITERATING) {
-		g_critical("recursive iteration on SDBM database \"%s\"",
+		s_critical("recursive iteration on SDBM database \"%s\"",
 			sdbm_name(db));
 	}
 
+#ifdef THREADS
+	if G_UNLIKELY(db->lock != NULL) {
+		uint stid = thread_small_id();
+
+		if G_UNLIKELY(db->iterid != THREAD_INVALID_ID && db->iterid != stid) {
+			s_critical("%s(): concurrent iteration on SDBM database \"%s\" "
+				"with %s from %s",
+				G_STRFUNC, sdbm_name(db),
+				thread_id_name(db->iterid), thread_name());
+			errno = EPERM;
+			value = nullitem;
+			goto done;
+		}
+
+		db->iterid = stid;
+	}
+#endif	/* THREADS */
+
 	db->flags |= DBM_ITERATING;
 	db->pagtail = lseek(db->pagf, 0L, SEEK_END);
 
@@ -1330,8 +1601,10 @@ sdbm_firstkey(DBM *db)
 	}
 #endif	/* LRU */
 
-	if G_UNLIKELY(db->pagtail < 0)
-		return iteration_done(db, FALSE);
+	if G_UNLIKELY(db->pagtail < 0) {
+		value = iteration_done(db, FALSE);
+		goto done;
+	}
 
 	/*
 	 * Start at page 0, skipping any page we can't read.
@@ -1347,7 +1620,10 @@ sdbm_firstkey(DBM *db)
 		/* Skip faulty page */
 	}
 
-	return getnext(db);
+	value = getnext(db);
+
+done:
+	sdbm_return_datum(db, value);
 }
 
 /**
@@ -1358,6 +1634,8 @@ sdbm_firstkey_safe(DBM *db)
 {
 	if (db != NULL) {
 		sdbm_check(db);
+
+		sdbm_synchronize(db);
 		db->flags |= DBM_KEYCHECK;
 
 		/*
@@ -1366,9 +1644,10 @@ sdbm_firstkey_safe(DBM *db)
 		 */
 
 		if G_UNLIKELY(db->flags & DBM_RDONLY) {
-			g_critical("%s() called on read-only SDBM database \"%s\"",
+			s_critical("%s() called on read-only SDBM database \"%s\"",
 				G_STRFUNC, sdbm_name(db));
 		}
+		sdbm_unsynchronize(db);
 	}
 	return sdbm_firstkey(db);
 }
@@ -1376,6 +1655,8 @@ sdbm_firstkey_safe(DBM *db)
 datum
 sdbm_nextkey(DBM *db)
 {
+	datum value;
+
 	if G_UNLIKELY(db == NULL) {
 		errno = EINVAL;
 		return nullitem;
@@ -1383,19 +1664,45 @@ sdbm_nextkey(DBM *db)
 
 	sdbm_check(db);
 
+	sdbm_synchronize(db);
+
 	if G_UNLIKELY(db->flags & DBM_BROKEN) {
 		errno = ESTALE;
-		return nullitem;
+		value = nullitem;
+		goto done;
 	}
 
 	if G_UNLIKELY(!(db->flags & DBM_ITERATING)) {
-		g_critical("%s() called outside of any key iteration over SDBM \"%s\"",
+		s_critical("%s() called outside of any key iteration over SDBM \"%s\"",
 			G_STRFUNC, sdbm_name(db));
 		errno = ENOENT;
-		return nullitem;
+		value = nullitem;
+		goto done;
 	}
 
-	return getnext(db);
+#ifdef THREADS
+	if G_UNLIKELY(db->lock != NULL) {
+		uint stid = thread_small_id();
+
+		/* Since DBM_ITERATING is set... */
+		g_soft_assert(db->iterid != THREAD_INVALID_ID);
+
+		if G_UNLIKELY(db->iterid != stid) {
+			s_critical("%s(): concurrent iteration on SDBM database \"%s\" "
+				"with %s from %s",
+				G_STRFUNC, sdbm_name(db),
+				thread_id_name(db->iterid), thread_name());
+			errno = EPERM;
+			value = nullitem;
+			goto done;
+		}
+	}
+#endif	/* THREADS */
+
+	value = getnext(db);
+
+done:
+	sdbm_return_datum(db, value);
 }
 
 /**
@@ -1406,15 +1713,34 @@ sdbm_endkey(DBM *db)
 {
 	sdbm_check(db);
 
+	sdbm_synchronize(db);
+
 	/*
 	 * Loudly warn if this is called outside of an iteration.
 	 */
 
 	if G_UNLIKELY(!(db->flags & DBM_ITERATING)) {
-		g_critical("%s() called outside of any key iteration over SDBM \"%s\"",
+		s_critical("%s() called outside of any key iteration over SDBM \"%s\"",
 			G_STRFUNC, sdbm_name(db));
 	}
 
+#ifdef THREADS
+	if G_UNLIKELY(db->lock != NULL) {
+		uint stid = thread_small_id();
+
+		/* Since DBM_ITERATING is set... */
+		g_soft_assert(db->iterid != THREAD_INVALID_ID);
+
+		if G_UNLIKELY(db->iterid != stid) {
+			s_critical("%s(): concurrent iteration on SDBM database \"%s\" "
+				"with %s from %s",
+				G_STRFUNC, sdbm_name(db),
+				thread_id_name(db->iterid), thread_name());
+			goto done;
+		}
+	}
+#endif	/* THREADS */
+
 	/*
 	 * When starting an iteration with sdbm_firstkey_safe() and encountering
 	 * big keys or values, a checking context is allocated and it is only freed
@@ -1422,6 +1748,9 @@ sdbm_endkey(DBM *db)
 	 */
 
 	(void) iteration_done(db, FALSE);		/* Iteration was interrupted */
+
+done:
+	sdbm_return_void(db);
 }
 
 /**
@@ -1492,6 +1821,8 @@ validpage(DBM *db, long pagb)
 	int removed = 0;
 	int corrupted = 0;
 
+	assert_sdbm_locked(db);
+
 	n = ino[0];
 
 	for (i = n - 1; i > 0; i -= 2) {
@@ -1509,7 +1840,7 @@ validpage(DBM *db, long pagb)
 				removed++;
 			} else {
 				/* Can happen on I/O error with big keys */
-				g_warning("sdbm: \"%s\": cannot remove key #%d/%d "
+				s_warning("sdbm: \"%s\": cannot remove key #%d/%d "
 					"not belonging to page #%ld",
 					sdbm_name(db), k, n / 2, pagb);
 			}
@@ -1518,7 +1849,7 @@ validpage(DBM *db, long pagb)
 			if (delipair(db, pag, i, FALSE)) {
 				corrupted++;
 			} else {
-				g_warning("sdbm: \"%s\": cannot remove corrupted entry #%d/%d "
+				s_warning("sdbm: \"%s\": cannot remove corrupted entry #%d/%d "
 					"in page #%ld",
 					sdbm_name(db), k, n / 2, pagb);
 			}
@@ -1528,15 +1859,15 @@ validpage(DBM *db, long pagb)
 	if G_UNLIKELY(removed > 0 || corrupted > 0) {
 		if (removed > 0) {
 			db->removed_keys += removed;
-			g_warning("sdbm: \"%s\": removed %d/%d key%s "
+			s_warning("sdbm: \"%s\": removed %d/%d key%s "
 				"not belonging to page #%ld", sdbm_name(db),
-				removed, n / 2, 1 == removed ? "" : "s", pagb);
+				removed, n / 2, plural(removed), pagb);
 		}
 		if (corrupted > 0) {
 			db->removed_keys += corrupted;
-			g_warning("sdbm: \"%s\": removed %d/%d corrupted entr%s "
+			s_warning("sdbm: \"%s\": removed %d/%d corrupted entr%s "
 				"on page #%ld", sdbm_name(db),
-				corrupted, n / 2, 1 == corrupted ? "y" : "ies", pagb);
+				corrupted, n / 2, plural_y(corrupted), pagb);
 		}
 #ifdef LRU
 		(void) force_flush_pagbuf(db, !db->is_volatile);
@@ -1549,6 +1880,8 @@ validpage(DBM *db, long pagb)
 static bool
 fetch_dirbuf(DBM *db, long dirb)
 {
+	assert_sdbm_locked(db);
+
 	db->dirfetch++;
 
 	if (dirb != db->dirbno) {
@@ -1562,7 +1895,7 @@ fetch_dirbuf(DBM *db, long dirb)
 		db->dirread++;
 		got = compat_pread(db->dirf, db->dirbuf, DBM_DBLKSIZ, OFF_DIR(dirb));
 		if G_UNLIKELY(got < 0) {
-			g_critical("sdbm: \"%s\": could not read dir page #%ld: %m",
+			s_critical("sdbm: \"%s\": could not read dir page #%ld: %m",
 				sdbm_name(db), dirb);
 			ioerr(db, FALSE);
 			return FALSE;
@@ -1601,6 +1934,8 @@ setdbit(DBM *db, long int dbit)
 	long c;
 	long dirb;
 
+	assert_sdbm_locked(db);
+
 	c = dbit / BYTESIZ;
 	dirb = c / DBM_DBLKSIZ;
 
@@ -1638,6 +1973,8 @@ getnext(DBM *db)
 {
 	datum key;
 
+	assert_sdbm_locked(db);
+
 	/*
 	 * During a traversal, no modification should be done on the database,
 	 * so the current page number must be the same as before.  The only
@@ -1685,22 +2022,27 @@ getnext(DBM *db)
 int
 sdbm_deletekey(DBM *db)
 {
+	int status = -1;
+
 	if G_UNLIKELY(db == NULL) {
 		errno = EINVAL;
 		return -1;
 	}
 	sdbm_check(db);
+
+	sdbm_synchronize(db);
+
 	if G_UNLIKELY(db->flags & DBM_RDONLY) {
 		errno = EPERM;
-		return -1;
+		goto done;
 	}
 	if G_UNLIKELY(db->flags & DBM_IOERR_W) {
 		errno = EIO;
-		return -1;
+		goto done;
 	}
 	if G_UNLIKELY(db->flags & DBM_BROKEN) {
 		errno = ESTALE;
-		return -1;
+		goto done;
 	}
 
 	/*
@@ -1708,18 +2050,36 @@ sdbm_deletekey(DBM *db)
 	 */
 
 	if G_UNLIKELY(!(db->flags & DBM_ITERATING)) {
-		g_critical("%s() called outside of any key iteration over SDBM \"%s\"",
+		s_critical("%s() called outside of any key iteration over SDBM \"%s\"",
 			G_STRFUNC, sdbm_name(db));
 		goto no_entry;
 	}
 
+#ifdef THREADS
+	if G_UNLIKELY(db->lock != NULL) {
+		uint stid = thread_small_id();
+
+		/* Since DBM_ITERATING is set... */
+		g_soft_assert(db->iterid != THREAD_INVALID_ID);
+
+		if G_UNLIKELY(db->iterid != stid) {
+			s_critical("%s(): concurrent iteration on SDBM database \"%s\" "
+				"with %s from %s",
+				G_STRFUNC, sdbm_name(db),
+				thread_id_name(db->iterid), thread_name());
+			errno = EPERM;
+			goto done;
+		}
+	}
+#endif	/* THREADS */
+
 	g_assert(db->pagbno == db->blkptr);	/* No page change since last time */
 
 	if G_UNLIKELY(0 == db->keyptr)
 		goto no_entry;
 
 	if G_UNLIKELY(!delnpair(db, db->pagbuf, db->keyptr))
-		return -1;
+		goto done;
 
 	db->keyptr--;
 
@@ -1728,13 +2088,16 @@ sdbm_deletekey(DBM *db)
 	 */
 
 	if G_UNLIKELY(!flush_pagbuf(db))
-		return -1;
+		goto done;
 
-	return 0;
+	status = 0;
+
+done:
+	sdbm_return(db, status);
 
 no_entry:
 	errno = ENOENT;
-	return -1;
+	goto done;
 }
 
 /**
@@ -1753,16 +2116,37 @@ sdbm_value(DBM *db)
 
 	sdbm_check(db);
 
+	sdbm_synchronize(db);
+
 	/*
 	 * Loudly warn if this is called outside of an iteration.
 	 */
 
 	if G_UNLIKELY(!(db->flags & DBM_ITERATING)) {
-		g_critical("%s() called outside of any key iteration over SDBM \"%s\"",
+		s_critical("%s() called outside of any key iteration over SDBM \"%s\"",
 			G_STRFUNC, sdbm_name(db));
 		goto no_entry;
 	}
 
+#ifdef THREADS
+	if G_UNLIKELY(db->lock != NULL) {
+		uint stid = thread_small_id();
+
+		/* Since DBM_ITERATING is set... */
+		g_soft_assert(db->iterid != THREAD_INVALID_ID);
+
+		if G_UNLIKELY(db->iterid != stid) {
+			s_critical("%s(): concurrent iteration on SDBM database \"%s\" "
+				"with %s from %s",
+				G_STRFUNC, sdbm_name(db),
+				thread_id_name(db->iterid), thread_name());
+			errno = EPERM;
+			val = nullitem;
+			goto done;
+		}
+	}
+#endif	/* THREADS */
+
 	g_assert(db->pagbno == db->blkptr);	/* No page change since last time */
 
 	if G_UNLIKELY(0 == db->keyptr)
@@ -1772,11 +2156,13 @@ sdbm_value(DBM *db)
 	if G_UNLIKELY(NULL == val.dptr)
 		goto no_entry;
 
-	return val;
+done:
+	sdbm_return_datum(db, val);
 
 no_entry:
 	errno = ENOENT;
-	return nullitem;
+	val = nullitem;
+	goto done;
 }
 
 /**
@@ -1793,19 +2179,26 @@ sdbm_sync(DBM *db)
 
 	sdbm_check(db);
 
+	sdbm_synchronize(db);
+
 	if G_UNLIKELY(db->flags & DBM_BROKEN) {
 		errno = ESTALE;
-		return -1;
+		npag = (ssize_t) -1;
+		goto done;
 	}
 
 #ifdef LRU
 	npag = flush_dirtypag(db);
-	if G_UNLIKELY(-1 == npag)
-		return -1;
+	if G_UNLIKELY(-1 == npag) {
+		npag = (ssize_t) -1;
+		goto done;
+	}
 
 	if (db->dirbuf_dirty) {
-		if G_UNLIKELY(!flush_dirbuf(db))
-			return -1;
+		if G_UNLIKELY(!flush_dirbuf(db)) {
+			npag = (ssize_t) -1;
+			goto done;
+		}
 		npag++;
 	}
 #else
@@ -1817,7 +2210,8 @@ sdbm_sync(DBM *db)
 		npag++;
 #endif
 
-	return npag;
+done:
+	sdbm_return(db, npag);
 }
 
 /**
@@ -1833,19 +2227,22 @@ sdbm_shrink(DBM *db)
 	filesize_t paglen;
 	filestat_t buf;
 	filesize_t offset;
+	bool status;
 
 	sdbm_check(db);
 
+	sdbm_synchronize(db);
+
 	if G_UNLIKELY(db->flags & DBM_BROKEN) {
 		errno = ESTALE;
-		return -1;
+		goto error;
 	}
 
 	if G_UNLIKELY(-1 == fstat(db->pagf, &buf))
-		return FALSE;
+		goto error;
 
 	if G_UNLIKELY(db->flags & DBM_RDONLY) {
-		g_critical("%s() called on read-only SDBM database \"%s\"",
+		s_critical("%s() called on read-only SDBM database \"%s\"",
 			G_STRFUNC, sdbm_name(db));
 	}
 
@@ -1899,7 +2296,7 @@ sdbm_shrink(DBM *db)
 
 	if (offset < paglen) {
 		if (-1 == ftruncate(db->pagf, offset))
-			return FALSE;
+			goto error;
 #ifdef LRU
 		lru_discard(db, truncate_bno);
 #endif
@@ -1927,7 +2324,7 @@ sdbm_shrink(DBM *db)
 		filesize = MAX(filesize, DBM_DBLKSIZ);	/* Ensure 1 block at least */
 
 		if G_UNLIKELY(-1 == fstat(db->dirf, &buf))
-			return FALSE;
+			goto error;
 
 		/*
 		 * Try to not change the mtime of the index if we don't have to.
@@ -1938,7 +2335,7 @@ sdbm_shrink(DBM *db)
 
 		if (filesize < buf.st_size) {
 			if G_UNLIKELY(-1 == ftruncate(db->dirf, filesize))
-				return FALSE;
+				goto error;
 			db->maxbno = filesize * BYTESIZ;
 		}
 
@@ -1952,7 +2349,7 @@ sdbm_shrink(DBM *db)
 			db->dirbno = -1;	/* Discard since after our truncation point */
 
 		if G_UNLIKELY(!fetch_dirbuf(db, dirb))
-			return FALSE;
+			goto error;
 
 		g_assert(filesize - maxsize < DBM_DBLKSIZ);
 
@@ -1990,16 +2387,23 @@ sdbm_shrink(DBM *db)
 	} else
 #endif
 	if G_UNLIKELY(!flush_dirbuf(db))
-		return FALSE;
+		goto error;
 
 no_idx_change:
 
 #ifdef BIGDATA
 	if G_UNLIKELY(!big_shrink(db))
-		return FALSE;
+		goto error;
 #endif
 
-	return TRUE;
+	status = TRUE;
+
+done:
+	sdbm_return(db, status);
+
+error:
+	status = FALSE;
+	goto done;
 }
 
 /**
@@ -2021,7 +2425,7 @@ int
 sdbm_rename_files(DBM *db,
 	const char *dirname, const char *pagname, const char *datname)
 {
-	int openflags, error = 0;
+	int openflags, error = 0, status;
 	bool dat_opened, dat_reopened;
 
 	if G_UNLIKELY(db == NULL) {
@@ -2030,15 +2434,17 @@ sdbm_rename_files(DBM *db,
 	}
 	sdbm_check(db);
 
+	sdbm_synchronize(db);
+
 	if G_UNLIKELY(db->flags & DBM_BROKEN) {
 		errno = ESTALE;
-		return -1;
+		goto error;
 	}
 
 #ifdef BIGDATA
 	if (NULL == datname && NULL != db->datname) {
 		errno = EINVAL;
-		return -1;
+		goto error;
 	}
 #endif
 
@@ -2067,17 +2473,17 @@ sdbm_rename_files(DBM *db,
 
 	if (-1 == rename(db->dirname, dirname)) {
 		error = errno;
-		g_critical("sdbm: \"%s\": cannot rename \"%s\" as \"%s\": %m",
+		s_critical("sdbm: \"%s\": cannot rename \"%s\" as \"%s\": %m",
 			sdbm_name(db), db->dirname, dirname);
 		goto emergency_restore;
 	}
 
 	if (-1 == rename(db->pagname, pagname)) {
 		error = errno;
-		g_critical("sdbm: \"%s\": cannot rename \"%s\" as \"%s\": %m",
+		s_critical("sdbm: \"%s\": cannot rename \"%s\" as \"%s\": %m",
 			sdbm_name(db), db->pagname, pagname);
 		if (-1 == rename(dirname, db->dirname)) {
-			g_warning("sdbm: \"%s\": cannot rename \"%s\" back to \"%s\": %m",
+			s_warning("sdbm: \"%s\": cannot rename \"%s\" back to \"%s\": %m",
 				sdbm_name(db), dirname, db->dirname);
 			db->flags |= DBM_BROKEN;
 		}
@@ -2089,15 +2495,15 @@ sdbm_rename_files(DBM *db,
 
 	if (-1 == rename(db->datname, datname)) {
 		error = errno;
-		g_critical("sdbm: \"%s\": cannot rename \"%s\" as \"%s\": %m",
+		s_critical("sdbm: \"%s\": cannot rename \"%s\" as \"%s\": %m",
 			sdbm_name(db), db->datname, datname);
 		if (-1 == rename(dirname, db->dirname)) {
-			g_warning("sdbm: \"%s\": cannot rename \"%s\" back to \"%s\": %m",
+			s_warning("sdbm: \"%s\": cannot rename \"%s\" back to \"%s\": %m",
 				sdbm_name(db), dirname, db->dirname);
 			db->flags |= DBM_BROKEN;
 		}
 		if (-1 == rename(pagname, db->pagname)) {
-			g_warning("sdbm: \"%s\": cannot rename \"%s\" back to \"%s\": %m",
+			s_warning("sdbm: \"%s\": cannot rename \"%s\" back to \"%s\": %m",
 				sdbm_name(db), pagname, db->pagname);
 			db->flags |= DBM_BROKEN;
 		}
@@ -2152,13 +2558,20 @@ emergency_restore:
 done:
 	if (error != 0) {
 		errno = error;
-		g_carp("sdbm: \"%s\": renaming operation %s: %m",
+		s_carp("sdbm: \"%s\": renaming operation %s: %m",
 			sdbm_name(db),
 			(db->flags & DBM_BROKEN) ? "broke database" : "failed");
-		return -1;
+		status = -1;
+	} else {
+		status = 0;
 	}
 
-	return 0;
+finished:
+	sdbm_return(db, status);
+
+error:
+	status = -1;
+	goto finished;
 }
 
 /**
@@ -2182,19 +2595,21 @@ sdbm_rename(DBM *db, const char *base)
 
 	if G_UNLIKELY(db == NULL) {
 		errno = EINVAL;
-		goto error;
+		return -1;
 	}
 
 	if (base == NULL || '\0' == base[0]) {
 		errno = EINVAL;
-		goto error;
+		return -1;
 	}
 
 	sdbm_check(db);
 
+	sdbm_synchronize(db);
+
 	if G_UNLIKELY(db->flags & DBM_BROKEN) {
 		errno = ESTALE;
-		return -1;
+		goto error;
 	}
 
 #ifdef BIGDATA
@@ -2233,11 +2648,11 @@ error:
 	HFREE_NULL(datname);
 
 	if (result != 0 && !warned) {
-		g_critical("sdbm: \"%s\": renaming operation failed: %m",
+		s_critical("sdbm: \"%s\": renaming operation failed: %m",
 			sdbm_name(db));
 	}
 
-	return result;
+	sdbm_return(db, result);
 }
 
 /**
@@ -2252,28 +2667,30 @@ sdbm_rebuild(DBM *db)
 	DBM *ndb;
 	char ext[10];
 	char *dirname, *pagname, *datname;
-	int error = 0;
+	int error = 0, result;
 	long cache;
 	datum key;
 	unsigned items = 0, skipped = 0, duplicate = 0;
 
 	sdbm_check(db);
 
+	sdbm_synchronize(db);
+
 	if (sdbm_rdonly(db)) {
 		errno = EPERM;
-		return -1;
+		goto failed;
 	}
 	if (sdbm_error(db)) {
 		errno = EIO;		/* Already got an error reported */
-		return -1;
+		goto failed;
 	}
 	if (db->flags & DBM_ITERATING) {
 		errno = EBUSY;		/* Already iterating */
-		return -1;
+		goto failed;
 	}
 	if G_UNLIKELY(db->flags & DBM_BROKEN) {
 		errno = ESTALE;		/* Already broken handle */
-		return -1;
+		goto failed;
 	}
 
 	str_bprintf(ext, sizeof ext, ".%08x", random_u32());
@@ -2328,6 +2745,7 @@ sdbm_rebuild(DBM *db)
 				continue;
 			}
 			/* Other errors are fatal */
+			error = errno;
 			sdbm_endkey(db);		/* Finish iteration */
 			break;
 		}
@@ -2348,8 +2766,14 @@ sdbm_rebuild(DBM *db)
 	pagname = h_strdup(db->pagname);
 	datname = h_strdup(db->datname);
 
+#ifdef THREADS
+	ndb->lock = db->lock;
+#endif
 	sdbm_close_internal(db, TRUE, FALSE);		/* Keep object around */
 	*db = *ndb;									/* struct copy */
+#ifdef THREADS
+	ndb->lock = NULL;							/* was copied over */
+#endif
 	sdbm_free_null(&ndb);
 
 	/*
@@ -2373,7 +2797,7 @@ error:
 
 	if (0 != error) {
 		errno = error;
-		return -1;
+		goto failed;
 	}
 
 	/*
@@ -2385,13 +2809,20 @@ error:
 	 */
 
 	if (skipped != 0) {
-		g_critical("sdbm: \"%s\": had to skip %u/%u item%s (%u duplicate%s)"
+		s_critical("sdbm: \"%s\": had to skip %u/%u item%s (%u duplicate%s)"
 			" during rebuild",
 			sdbm_name(db), skipped, items, 1 == skipped ? "" : "s",
 			duplicate, 1 == duplicate ? "" : "s");
 	}
 
-	return 0;		/* OK, we rebuilt the database */
+	result = 0;		/* OK, we rebuilt the database */
+
+done:
+	sdbm_return(db, result);
+
+failed:
+	result = -1;
+	goto done;
 }
 
 /**
@@ -2402,25 +2833,30 @@ error:
 int
 sdbm_clear(DBM *db)
 {
+	int result;
+
 	if G_UNLIKELY(db == NULL) {
 		errno = EINVAL;
 		return -1;
 	}
 	sdbm_check(db);
+
+	sdbm_synchronize(db);
+
 	if G_UNLIKELY(db->flags & DBM_RDONLY) {
 		errno = EPERM;
-		return -1;
+		goto error;
 	}
 	if G_UNLIKELY(db->flags & DBM_BROKEN) {
 		errno = ESTALE;
-		return -1;
+		goto error;
 	}
 	if G_UNLIKELY(-1 == ftruncate(db->pagf, 0))
-		return -1;
+		goto error;
 	db->pagbno = -1;
 	db->pagtail = 0L;
 	if G_UNLIKELY(-1 == ftruncate(db->dirf, 0))
-		return -1;
+		goto error;
 	db->dirbno = -1;
 	db->maxbno = 0;
 	db->curbit = 0;
@@ -2433,9 +2869,16 @@ sdbm_clear(DBM *db)
 	sdbm_clearerr(db);
 #ifdef BIGDATA
 	if G_UNLIKELY(!big_clear(db))
-		return -1;
+		goto error;
 #endif
-	return 0;
+	result = 0;
+
+done:
+	sdbm_return(db, result);
+
+error:
+	result = -1;
+	goto done;
 }
 
 /**
@@ -2444,12 +2887,19 @@ sdbm_clear(DBM *db)
 long
 sdbm_get_cache(const DBM *db)
 {
+	long pages;
+
 	sdbm_check(db);
+
+	sdbm_synchronize(db);
+
 #ifdef LRU
-	return getcache(db);
+	pages = getcache(db);
 #else
-	return 0;
+	pages = 0;
 #endif
+
+	sdbm_return(db, pages);
 }
 
 /**
@@ -2458,16 +2908,23 @@ sdbm_get_cache(const DBM *db)
 int
 sdbm_set_cache(DBM *db, long pages)
 {
+	int result;
+
 	sdbm_check(db);
+
+	sdbm_synchronize(db);
+
 #ifdef LRU
 	if G_UNLIKELY(NULL == db->cache)
 		lru_init(db);
-	return setcache(db, pages);
+	result = setcache(db, pages);
 #else
 	(void) pages;
 	errno = ENOTSUP;
-	return -1;
+	result = -1;
 #endif
+
+	sdbm_return(db, result);
 }
 
 /**
@@ -2476,13 +2933,19 @@ sdbm_set_cache(DBM *db, long pages)
 bool
 sdbm_get_wdelay(const DBM *db)
 {
+	bool delayed;
+
 	sdbm_check(db);
 
+	sdbm_synchronize(db);
+
 #ifdef LRU
-	return getwdelay(db);
+	delayed = getwdelay(db);
 #else
-	return FALSE;
+	delayed = FALSE;
 #endif
+
+	sdbm_return(db, delayed);
 }
 
 /**
@@ -2491,14 +2954,21 @@ sdbm_get_wdelay(const DBM *db)
 int
 sdbm_set_wdelay(DBM *db, bool on)
 {
+	int result;
+
 	sdbm_check(db);
+
+	sdbm_synchronize(db);
+
 #ifdef LRU
-	return setwdelay(db, on);
+	result = setwdelay(db, on);
 #else
 	(void) on;
 	errno = ENOTSUP;
-	return -1;
+	result = -1;
 #endif
+
+	sdbm_return(db, result);
 }
 
 /**
@@ -2507,12 +2977,19 @@ sdbm_set_wdelay(DBM *db, bool on)
 bool
 sdbm_is_volatile(const DBM *db)
 {
+	bool result;
+
 	sdbm_check(db);
+
+	sdbm_synchronize(db);
+
 #ifdef LRU
-	return db->is_volatile;
+	result = db->is_volatile;
 #else
-	return FALSE;
+	result = FALSE;
 #endif
+
+	sdbm_return(db, result);
 }
 
 /**
@@ -2523,73 +3000,108 @@ sdbm_is_volatile(const DBM *db)
 int
 sdbm_set_volatile(DBM *db, bool yes)
 {
+	int result;
+
 	sdbm_check(db);
+
 #ifdef LRU
 	db->is_volatile = yes;
-	if (yes)
-		return setwdelay(db, TRUE);
+	result = yes ? setwdelay(db, TRUE) : 0;
 #else
 	(void) yes;
+	result = 0;
 #endif
-	return 0;
+
+	sdbm_return(db, result);
 }
 
 bool
 sdbm_rdonly(DBM *db)
 {
+	bool rdonly;
+
 	sdbm_check(db);
-	return 0 != (db->flags & DBM_RDONLY);
+
+	sdbm_synchronize(db);
+	rdonly = 0 != (db->flags & DBM_RDONLY);
+	sdbm_return(db, rdonly);
 }
 
 bool
 sdbm_error(DBM *db)
 {
+	bool error;
+
 	sdbm_check(db);
-	return 0 != (db->flags & (DBM_IOERR | DBM_IOERR_W));
+
+	sdbm_synchronize(db);
+	error = 0 != (db->flags & (DBM_IOERR | DBM_IOERR_W));
+	sdbm_return(db, error);
 }
 
 void
 sdbm_clearerr(DBM *db)
 {
 	sdbm_check(db);
+
+	sdbm_synchronize(db);
 	db->flags &= ~(DBM_IOERR | DBM_IOERR_W);
+	sdbm_unsynchronize(db);
 }
 
 int
 sdbm_dirfno(DBM *db)
 {
+	int fno;
+
 	sdbm_check(db);
 
-	if G_UNLIKELY(db->flags & DBM_BROKEN)
-		return -1;
+	sdbm_synchronize(db);
 
-	return db->dirf;
+	if G_UNLIKELY(db->flags & DBM_BROKEN)
+		fno = -1;
+	else
+		fno = db->dirf;
+
+	sdbm_return(db, fno);
 }
 
 int
 sdbm_pagfno(DBM *db)
 {
+	int fno;
+
 	sdbm_check(db);
 
-	if G_UNLIKELY(db->flags & DBM_BROKEN)
-		return -1;
+	sdbm_synchronize(db);
 
-	return db->pagf;
+	if G_UNLIKELY(db->flags & DBM_BROKEN)
+		fno = -1;
+	else
+		fno = db->pagf;
+
+	sdbm_return(db, fno);
 }
 
 int
 sdbm_datfno(DBM *db)
 {
+	int fno;
+
 	sdbm_check(db);
 
+	sdbm_synchronize(db);
+
 	if G_UNLIKELY(db->flags & DBM_BROKEN)
-		return -1;
+		fno = -1;
 
 #ifdef BIGDATA
-	return big_datfno(db);
+	fno = big_datfno(db);
 #else
-	return -1;
+	fno = -1;
 #endif
+
+	sdbm_return(db, fno);
 }
 
 /* vi: set ts=4 sw=4 cindent: */
