@@ -46,6 +46,7 @@
 #include "if/gui_property.h"
 #include "if/gui_property_priv.h"
 #include "if/bridge/ui2c.h"
+#include "if/core/sockets.h"		/* For SOCK_F_G2 */
 
 #include "lib/atoms.h"
 #include "lib/concat.h"
@@ -363,7 +364,7 @@ nodes_gui_update_node_flags(struct node_data *data, gnet_node_flags_t *flags)
 	concat_strings(data->flags, sizeof data->flags,
 		"<tt>", guc_node_flags_to_string(flags), "</tt>", (void *) 0);
 
-	ultra = NODE_P_ULTRA == flags->peermode;
+	ultra = NODE_P_ULTRA == flags->peermode || NODE_P_G2HUB == flags->peermode;
     data->fg = &(gtk_widget_get_style(GTK_WIDGET(treeview_nodes))
 					->fg[ultra ? GTK_STATE_NORMAL : GTK_STATE_INSENSITIVE]);
 }
@@ -886,7 +887,7 @@ nodes_gui_browse_selected_helper(GtkTreeModel *model,
 	info = guc_node_get_info(data->node_id);
 	if (!info->is_pseudo) {
 		search_gui_new_browse_host(NULL, info->gnet_addr, info->gnet_port,
-			&info->gnet_guid, NULL, 0);
+			&info->gnet_guid, NULL, info->is_g2 ? SOCK_F_G2 : 0);
 	}
 	guc_node_free_info(info);
 }
