@@ -44,6 +44,8 @@
 #include "if/bridge/ui2c.h"
 
 #include "lib/glib-missing.h"	/* For g_strlcpy() */
+#include "lib/halloc.h"
+#include "lib/pslist.h"
 #include "lib/str.h"
 #include "lib/override.h"		/* Must be the last header included */
 
@@ -153,7 +155,8 @@ on_button_config_remove_dir_clicked(GtkButton *unused_button,
 	GtkTreeModel *model;
 	GtkTreeIter iter;
 	GtkTreeSelection *selection;
-	str_t *s;
+	pslist_t *sl, *pl_dirs = NULL;
+	char *dirs;
 
 	(void) unused_button;
 	(void) unused_udata;
@@ -166,24 +169,26 @@ on_button_config_remove_dir_clicked(GtkButton *unused_button,
 
 	/* Regenerate the string property holding a list of paths */
 	selection = gtk_tree_view_get_selection(tv);
-	s = str_new(0);
 
 	do {
-		gchar *dir = NULL;
+		char *pathname = NULL;
 
 		/* Skip items selected for removal */
 		if (gtk_tree_selection_iter_is_selected(selection, &iter))
 			continue;
 
-		gtk_tree_model_get(model, &iter, 0, &dir, (-1));
-		if (str_len(s) > 0)
-			str_putc(s, ':');
-		str_cat(s, dir);
-		G_FREE_NULL(dir);
+		gtk_tree_model_get(model, &iter, 0, &pathname, (-1));
+		pl_dirs = pslist_prepend(pl_dirs, pathname);
 	} while (gtk_tree_model_iter_next(model, &iter));
 
-	gnet_prop_set_string(PROP_SHARED_DIRS_PATHS, str_2c(s));
-	str_destroy(s);
+	dirs = dirlist_to_string(pl_dirs);
+	gnet_prop_set_string(PROP_SHARED_DIRS_PATHS, dirs);
+	HFREE_NULL(dirs);
+
+	PSLIST_FOREACH(pl_dirs, sl) {
+		G_FREE_NULL(sl->data);
+	}
+	pslist_free_null(&pl_dirs);
 }
 #endif /* USE_GTK2 */
 
@@ -621,10 +626,10 @@ dbg_tree_init(void)
 	gui_signal_connect(tv, "cursor-changed", on_cursor_changed, tv);
 }
 
-void
-dbg_property_show_list(const GSList *props)
+static void
+dbg_property_show_list(const pslist_t *props)
 {
-	const GSList *sl;
+	const pslist_t *sl;
 	GtkTreeView *tv;
 	GtkListStore *store;
 	
@@ -656,7 +661,7 @@ dbg_property_show_list(const GSList *props)
 #endif /* USE_GTK2 */
 	}
 
-	for (sl = props; NULL != sl; sl = g_slist_next(sl)) {
+	PSLIST_FOREACH(props, sl) {
 		GtkTreeIter iter;
 		property_t prop;
 
@@ -748,17 +753,17 @@ dbg_property_set_row(GtkCList *clist, gint row, property_t prop)
 }
 
 static void
-dbg_property_show_list(const GSList *props)
+dbg_property_show_list(const pslist_t *props)
 {
 	GtkCList *clist;
-	const GSList *sl;
+	const pslist_t *sl;
 
    	clist = GTK_CLIST(gui_dlg_prefs_lookup("clist_dbg_property"));
 	gtk_clist_freeze(clist);
 	gtk_clist_clear(clist);
 
-	for (sl = props; NULL != sl; sl = g_slist_next(sl)) {
-		static const gchar * const titles[num_dbg_cols] = { "", "", "", "", };
+	PSLIST_FOREACH(props, sl) {
+		static const char * const titles[num_dbg_cols] = { "", "", "", "", };
 		property_t prop = GPOINTER_TO_UINT(sl->data);
 		gint row;
 		
@@ -905,7 +910,7 @@ on_entry_dbg_property_pattern_activate(GtkEditable *unused_editable,
 	g_strstrip(text);
 
 	if (0 != strcmp(text, old_pattern)) {
-		GSList *props;
+		pslist_t *props;
 
 		g_strlcpy(old_pattern, text, sizeof old_pattern);
 		props = gnet_prop_get_by_regex(text, NULL);
@@ -915,7 +920,7 @@ on_entry_dbg_property_pattern_activate(GtkEditable *unused_editable,
 		
 		dbg_property_show_list(props);
 		dbg_property_update_selection();
-		gm_slist_free_null(&props);
+		pslist_free_null(&props);
 	}
 	G_FREE_NULL(text);
 }
