@@ -52,6 +52,7 @@
 #include "lib/cq.h"
 #include "lib/getgateway.h"
 #include "lib/pmsg.h"
+#include "lib/stringify.h"
 #include "lib/tm.h"
 #include "lib/walloc.h"
 
@@ -125,7 +126,7 @@ natpmp_rpc_check(const struct natpmp_rpc * const rd)
 
 static unsigned natpmp_rpc_pending;
 
-static void natpmp_rpc_iterate(cqueue_t *unused_cq, void *obj);
+static void natpmp_rpc_iterate(cqueue_t *cq, void *obj);
 
 /**
  * Allocate a new NAT-PMP gateway.
@@ -462,10 +463,10 @@ error:
 		if (bstr_has_error(bs)) {
 			g_warning("NATPMP parsing error while processing discovery reply "
 				"(%zu byte%s): %s",
-				len, 1 == len ? "" : "s", bstr_error(bs));
+				len, plural(len), bstr_error(bs));
 		} else {
 			g_warning("NATPMP inconsistent discovery reply (%zu byte%s)",
-				len, 1 == len ? "" : "s");
+				len, plural(len));
 		}
 	}
 	bstr_free(&bs);
@@ -579,10 +580,10 @@ error:
 		if (bstr_has_error(bs)) {
 			g_warning("NATPMP parsing error while processing discovery reply "
 				"(%zu byte%s): %s",
-				len, 1 == len ? "" : "s", bstr_error(bs));
+				len, plural(len), bstr_error(bs));
 		} else {
 			g_warning("NATPMP inconsistent discovery reply (%zu byte%s)",
-				len, 1 == len ? "" : "s");
+				len, plural(len));
 		}
 	}
 	bstr_free(&bs);
@@ -604,7 +605,7 @@ natpmp_rpc_reply(enum urpc_ret type, host_addr_t addr, uint16 port,
 		g_debug("NATPMP %s for \"%s\" #%u (%lu byte%s) from %s",
 			URPC_TIMEOUT == type ? "timeout" : "got reply",
 			natpmp_op_to_string(rd->op), rd->count,
-			(unsigned long) len, 1 == len ? "" : "s",
+			(unsigned long) len, plural(len),
 			host_addr_port_to_string(addr, port));
 	}
 
@@ -616,7 +617,7 @@ natpmp_rpc_reply(enum urpc_ret type, host_addr_t addr, uint16 port,
 	 * sent the RPC.
 	 */
 
-	if (!host_addr_equal(addr, rd->gateway)) {
+	if (!host_addr_equiv(addr, rd->gateway)) {
 		if (GNET_PROPERTY(natpmp_debug)) {
 			g_warning("NATPMP discarding reply from %s (sent %s to %s)",
 				host_addr_port_to_string(addr, port),
